# Note: This file patches the ContactSensor class to support `filter_prim_paths_expr` at any level of the hierarchy

import isaaclab.sim as sim_utils
import omni.physics.tensors.impl.api as physx
import torch
from isaaclab.sensors import ContactSensor, ContactSensorCfg  # noqa: F401
from pxr import UsdPhysics

from srb.core.sensor import SensorBase

from pxr import PhysxSchema  # isort: skip


def __initialize_impl(self):
    SensorBase._initialize_impl(self)
    # create simulation view
    self._physics_sim_view = physx.create_simulation_view(self._backend)
    self._physics_sim_view.set_subspace_roots("/")
    # check that only rigid bodies are selected
    leaf_pattern = self.cfg.prim_path.rsplit("/", 1)[-1]
    template_prim_path = self._parent_prims[0].GetPath().pathString
    body_names = []
    for prim in sim_utils.find_matching_prims(template_prim_path + "/" + leaf_pattern):
        # check if prim has contact reporter API
        if prim.HasAPI(PhysxSchema.PhysxContactReportAPI):
            prim_path = prim.GetPath().pathString
            body_names.append(prim_path.rsplit("/", 1)[-1])
    # check that there is at least one body with contact reporter API
    if not body_names:
        raise RuntimeError(
            f"Sensor at path '{self.cfg.prim_path}' could not find any bodies with contact reporter API."
            "\nHINT: Make sure to enable 'activate_contact_sensors' in the corresponding asset spawn configuration."
        )

    # construct regex expression for the body names
    body_names_regex = r"(" + "|".join(body_names) + r")"
    body_names_regex = f"{self.cfg.prim_path.rsplit('/', 1)[0]}/{body_names_regex}"
    # convert regex expressions to glob expressions for PhysX
    body_names_glob = body_names_regex.replace(".*", "*")

    ### PATCH BEGINS HERE ###
    # filter_prim_paths_glob = [
    #     expr.replace(".*", "*") for expr in self.cfg.filter_prim_paths_expr
    # ]
    filter_prim_paths_glob = []
    for expr in self.cfg.filter_prim_paths_expr:
        queue = sim_utils.find_matching_prims(
            expr.replace(".*", "*").replace("/World/envs/env_*", "/World/envs/env_0")
        )
        while queue:
            child_prim = queue.pop(0)
            if child_prim.HasAPI(
                UsdPhysics.CollisionAPI  # type: ignore
            ) or child_prim.HasAPI(PhysxSchema.PhysxCollisionAPI):
                filter_prim_paths_glob.append(child_prim.GetPath().pathString)
            else:
                queue.extend(child_prim.GetChildren())
    if not filter_prim_paths_glob:
        filter_prim_paths_glob = [
            expr.replace(".*", "*") for expr in self.cfg.filter_prim_paths_expr
        ]
    ### PATCH ENDS HERE ###

    # create a rigid prim view for the sensor
    self._body_physx_view = self._physics_sim_view.create_rigid_body_view(
        body_names_glob
    )
    self._contact_physx_view = self._physics_sim_view.create_rigid_contact_view(
        body_names_glob, filter_patterns=filter_prim_paths_glob
    )
    # resolve the true count of bodies
    self._num_bodies = self.body_physx_view.count // self._num_envs
    # check that contact reporter succeeded
    if self._num_bodies != len(body_names):
        raise RuntimeError(
            "Failed to initialize contact reporter for specified bodies."
            f"\n\tInput prim path    : {self.cfg.prim_path}"
            f"\n\tResolved prim paths: {body_names_regex}"
        )

    # prepare data buffers
    self._data.net_forces_w = torch.zeros(
        self._num_envs, self._num_bodies, 3, device=self._device
    )
    # optional buffers
    # -- history of net forces
    if self.cfg.history_length > 0:
        self._data.net_forces_w_history = torch.zeros(
            self._num_envs,
            self.cfg.history_length,
            self._num_bodies,
            3,
            device=self._device,
        )
    else:
        self._data.net_forces_w_history = self._data.net_forces_w.unsqueeze(1)
    # -- pose of sensor origins
    if self.cfg.track_pose:
        self._data.pos_w = torch.zeros(
            self._num_envs, self._num_bodies, 3, device=self._device
        )
        self._data.quat_w = torch.zeros(
            self._num_envs, self._num_bodies, 4, device=self._device
        )
    # -- air/contact time between contacts
    if self.cfg.track_air_time:
        self._data.last_air_time = torch.zeros(
            self._num_envs, self._num_bodies, device=self._device
        )
        self._data.current_air_time = torch.zeros(
            self._num_envs, self._num_bodies, device=self._device
        )
        self._data.last_contact_time = torch.zeros(
            self._num_envs, self._num_bodies, device=self._device
        )
        self._data.current_contact_time = torch.zeros(
            self._num_envs, self._num_bodies, device=self._device
        )
    # force matrix: (num_envs, num_bodies, num_filter_shapes, 3)
    if len(self.cfg.filter_prim_paths_expr) != 0:
        num_filters = self.contact_physx_view.filter_count
        self._data.force_matrix_w = torch.zeros(
            self._num_envs, self._num_bodies, num_filters, 3, device=self._device
        )


ContactSensor._initialize_impl = __initialize_impl



from isaaclab.envs.mdp.actions import *  # noqa: F403
from isaaclab.envs.mdp.actions.task_space_actions import *  # noqa: F403

from .action_group import ActionGroup, ActionGroupRegistry  # noqa: F401
from .group import *  # noqa: F403
from .term import *  # noqa: F403



from srb.core.action import ActionGroup, WheeledDriveActionCfg, WheeledDriveActionGroup
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import ArticulationCfg, Frame, Transform, WheeledRobot
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import deg_to_rad, rpy_to_quat
from srb.utils.path import SRB_ASSETS_DIR_SRB_ROBOT


class MarsRover(WheeledRobot):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/mars_rover",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("PRIVATE").as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.02, rest_offset=0.005
            ),
            rigid_props=RigidBodyPropertiesCfg(
                max_linear_velocity=1.5,
                max_angular_velocity=1000.0,
                max_depenetration_velocity=1.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=16,
                solver_velocity_iteration_count=4,
            ),
        ),
        actuators={
            "base_steering": ImplicitActuatorCfg(
                joint_names_expr=[".*Steer.*"],
                velocity_limit=6.0,
                effort_limit=12.0,
                stiffness=8000.0,
                damping=1000.0,
            ),
            "base_drive": ImplicitActuatorCfg(
                joint_names_expr=[".*Drive.*"],
                velocity_limit=6.0,
                effort_limit=12.0,
                stiffness=100.0,
                damping=4000.0,
            ),
            "passive_joints_boogie": ImplicitActuatorCfg(
                joint_names_expr=[".*RevoluteJoint"],
                velocity_limit=15.0,
                effort_limit=0.0,
                stiffness=0.0,
                damping=0.0,
            ),
            "passive_joints_rocker": ImplicitActuatorCfg(
                joint_names_expr=[".*Rocker.*"],
                velocity_limit=15.0,
                effort_limit=0.0,
                stiffness=0.0,
                damping=0.0,
            ),
            "passive_joints_differential": ImplicitActuatorCfg(
                joint_names_expr=[".*Differential.*"],
                velocity_limit=15.0,
                effort_limit=0.0,
                stiffness=0.0,
                damping=0.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = WheeledDriveActionGroup(
        WheeledDriveActionCfg(
            asset_name="robot",
            wheelbase=(0.849, 0.77),
            wheelbase_mid=0.894,
            wheel_radius=0.1,
            steering_joint_names=[
                "FL_Steer_Revolute",
                "FR_Steer_Revolute",
                "RL_Steer_Revolute",
                "RR_Steer_Revolute",
            ],
            drive_joint_names=[
                "FL_Drive_Continous",
                "FR_Drive_Continous",
                "CL_Drive_Continous",
                "CR_Drive_Continous",
                "RL_Drive_Continous",
                "RR_Drive_Continous",
            ],
            scale_linear=0.5,
            scale_angular=deg_to_rad(45),
        )
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="Body")
    frame_payload_mount: Frame = Frame(
        prim_relpath="Body",
        offset=Transform(
            pos=(-0.22, 0.0, 0.23),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="Body",
        offset=Transform(
            pos=(0.25, 0.0, 0.1),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_front_camera: Frame = Frame(
        prim_relpath="Body/camera_front",
        offset=Transform(
            pos=(-0.7675, 0.0, 1.9793),
            rot=rpy_to_quat(0.0, 15.0, -90.0),
        ),
    )



from srb.core.asset import Frame, Object, RigidObjectCfg, Transform
from srb.core.sim import (
    CollisionPropertiesCfg,
    MassPropertiesCfg,
    MeshCollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.path import SRB_ASSETS_DIR_SRB_OBJECT


class BoltM8(Object):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/peg",
        spawn=UsdFileCfg(
            usd_path=(
                SRB_ASSETS_DIR_SRB_OBJECT.joinpath("bolt_and_nut")
                .joinpath("bolt_hex_m8.usdz")
                .as_posix()
            ),
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="sdf",
                sdf_resolution=512,
            ),
            rigid_props=RigidBodyPropertiesCfg(),
            mass_props=MassPropertiesCfg(mass=0.1),
        ),
    )

    ## Frames
    frame_driver_slot: Frame = Frame(offset=Transform(pos=(0.0, 0.0, 0.01905)))


class NutM8(Object):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/hole",
        spawn=UsdFileCfg(
            usd_path=(
                SRB_ASSETS_DIR_SRB_OBJECT.joinpath("bolt_and_nut")
                .joinpath("nut_hex_m8.usdz")
                .as_posix()
            ),
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="sdf",
                sdf_resolution=512,
            ),
            rigid_props=RigidBodyPropertiesCfg(kinematic_enabled=True),
            mass_props=MassPropertiesCfg(mass=0.05),
        ),
    )



import threading
from collections.abc import Callable
from typing import TYPE_CHECKING, Sequence

import numpy
import torch

from srb.core.action import ActionGroup
from srb.interfaces.enums import TeleopDeviceType
from srb.interfaces.teleop import DeviceBase, Se3Gamepad
from srb.interfaces.teleop.keyboard import KeyboardTeleopInterface
from srb.interfaces.teleop.spacemouse import SpacemouseTeleopInterface

if TYPE_CHECKING:
    from rclpy.node import Node


class CombinedTeleopInterface(DeviceBase):
    def __init__(
        self,
        devices: Sequence[TeleopDeviceType],
        node: "Node | None" = None,
        pos_sensitivity: float = 1.0,
        rot_sensitivity: float = 3.1415927,
        actions: ActionGroup | None = None,
    ):
        if not node and (
            TeleopDeviceType.ROS in devices or TeleopDeviceType.HAPTIC in devices
        ):
            from srb.utils.ros import enable_ros2_bridge

            enable_ros2_bridge()
            import rclpy
            from rclpy.node import Node

            rclpy.init()
            self._node = Node("sim", namespace="srb", start_parameter_services=False)  # type: ignore
        else:
            self._node = node

        self._actions = actions
        self.interfaces = []
        self.ft_feedback_interfaces = []
        for device in devices:
            match device.lower():
                case TeleopDeviceType.KEYBOARD:
                    self.interfaces.append(
                        KeyboardTeleopInterface(
                            pos_sensitivity=0.6 * pos_sensitivity,
                            rot_sensitivity=0.4 * rot_sensitivity,
                        )
                    )
                case TeleopDeviceType.SPACEMOUSE:
                    self.interfaces.append(
                        SpacemouseTeleopInterface(
                            pos_sensitivity=2.0 * pos_sensitivity,
                            rot_sensitivity=0.8 * rot_sensitivity,
                        )
                    )
                case TeleopDeviceType.GAMEPAD:
                    self.interfaces.append(
                        Se3Gamepad(
                            pos_sensitivity=10.0 * pos_sensitivity,
                            rot_sensitivity=8.0 * rot_sensitivity,
                        )
                    )
                case TeleopDeviceType.ROS:
                    from srb.interfaces.teleop.ros import ROSTeleopInterface

                    self.interfaces.append(
                        ROSTeleopInterface(
                            node=self._node,
                            # Note: Gains assume that ROS messages originate from a gamepad
                            pos_sensitivity=10.0 * pos_sensitivity,
                            rot_sensitivity=8.0 * rot_sensitivity,
                        )
                    )
                case TeleopDeviceType.HAPTIC:
                    from srb.interfaces.teleop.haptic import HapticROSTeleopInterface

                    interface = HapticROSTeleopInterface(
                        node=self._node,
                        pos_sensitivity=8.0 * pos_sensitivity,
                        rot_sensitivity=1.0 * rot_sensitivity,
                    )
                    self.interfaces.append(interface)
                    self.ft_feedback_interfaces.append(interface)
                case _:
                    raise ValueError(f"Invalid device interface '{device}'.")

            self.gain = 1.0

            def cb_gain_decrease():
                self.gain *= 0.75
                print(f"Gain: {self.gain}")

            self.add_callback("O", cb_gain_decrease)

            def cb_gain_increase():
                self.gain *= 1.25
                print(f"Gain: {self.gain}")

            self.add_callback("P", cb_gain_increase)

        # Run a thread for listening to device
        if not node and self._node is not None:
            from rclpy.executors import MultiThreadedExecutor

            self._executor = MultiThreadedExecutor(num_threads=2)
            self._executor.add_node(self._node)
            self._thread = threading.Thread(target=self._executor.spin)
            self._thread.daemon = True
            self._thread.start()

    def __del__(self):
        for interface in self.interfaces:
            interface.__del__()

    def __str__(self) -> str:
        from srb.interfaces.teleop.keyboard import KeyboardTeleopInterface

        msg = "Combined Interface\n"
        msg += f"Devices: {', '.join([interface.__class__.__name__ for interface in self.interfaces])}\n"

        for interface in self.interfaces:
            if (
                isinstance(interface, KeyboardTeleopInterface)
                and self._actions is not None
            ):
                msg += self._keyboard_control_scheme()
                continue
            msg += "\n"
            msg += interface.__str__()

        return msg

    """
    Operations
    """

    def reset(self):
        for interface in self.interfaces:
            interface.reset()

        self._close_gripper = False
        self._prev_gripper_cmds = [False] * len(self.interfaces)

    def add_callback(self, key: str, func: Callable):
        for interface in self.interfaces:
            if isinstance(interface, KeyboardTeleopInterface):
                interface.add_callback(key=key, func=func)
            if isinstance(interface, SpacemouseTeleopInterface) and key in [
                "L",
                "R",
                "LR",
            ]:
                interface.add_callback(key=key, func=func)

    def advance(self) -> tuple[numpy.ndarray, bool]:
        raw_actions = [interface.advance() for interface in self.interfaces]

        twist = self.gain * numpy.sum(
            numpy.stack([a[0] for a in raw_actions], axis=0), axis=0
        )

        for i, prev_gripper_cmd in enumerate(self._prev_gripper_cmds):
            if prev_gripper_cmd != raw_actions[i][1]:
                self._close_gripper = not self._close_gripper
                break
        self._prev_gripper_cmds = [a[1] for a in raw_actions]

        return twist, self._close_gripper

    def set_ft_feedback(self, ft_feedback: numpy.ndarray | torch.Tensor):
        for interface in self.ft_feedback_interfaces:
            interface.set_ft_feedback(ft_feedback)

    @staticmethod
    def _keyboard_control_scheme() -> str:
        return """
+------------------------------------------------+
|  Keyboard Scheme (focus the Isaac Sim window)  |
+------------------------------------------------+
+------------------------------------------------+
| Reset: [ L ]                                   |
| Decrease Gain [ O ]   | Increase Gain: [ P ]   |
| Event: [ R / K ]                               |
+------------------------------------------------+
| Translation                                    |
|             [ W ] (+X)            [ Q ] (+Z)   |
|               ↑                     ↑          |
|               |                     |          |
|  (-Y) [ A ] ← + → [ D ] (+Y)        +          |
|               |                     |          |
|               ↓                     ↓          |
|             [ S ] (-X)            [ E ] (-Z)   |
|------------------------------------------------|
| Rotation                                       |
|       [ Z ] ←--------(±X)--------→ [ X ]       |
|                                                |
|       [ T ] ↻--------(±Y)--------↺ [ G ]       |
|                                                |
|       [ C ] ↺--------(±Z)--------↻ [ V ]       |
+------------------------------------------------+
        """



from typing import Callable

from isaaclab.managers import ActionManager as __ActionManager

from srb.core.action import ActionTerm, ActionTermCfg


class ActionManager(__ActionManager):
    def _prepare_terms(self):
        # Create buffers to parse and store terms
        self._term_names: list[str] = []
        self._terms: dict[str, ActionTerm] = {}

        # Check if config is dict already
        if isinstance(self.cfg, dict):
            cfg_items = self.cfg.items()
        else:
            cfg_items = self.cfg.__dict__.items()
        # Parse action terms from the config
        for term_name, term_cfg in cfg_items:
            # Check if term config is None
            if term_cfg is None or isinstance(term_cfg, Callable):
                continue
            # Check valid type
            if not isinstance(term_cfg, ActionTermCfg):
                raise TypeError(
                    f"Configuration for the term '{term_name}' is not of type ActionTermCfg."
                    f" Received: '{type(term_cfg)}'."
                )
            # Create the action term
            term = term_cfg.class_type(term_cfg, self._env)
            # Sanity check if term is valid type
            if not isinstance(term, ActionTerm):
                raise TypeError(
                    f"Returned object for the term '{term_name}' is not of type ActionType."
                )
            # Add term name and parameters
            self._term_names.append(term_name)
            self._terms[term_name] = term



from typing import TYPE_CHECKING, Dict, Sequence, Union

import orjson as json

from srb.utils.isaacsim import is_isaacsim_initialized
from srb.utils.path import (
    SRB_CACHE_PATH,
    SRB_ENV_CACHE_PATH,
    SRB_OBJECT_CACHE_PATH,
    SRB_ROBOT_CACHE_PATH,
    SRB_SCENERY_CACHE_PATH,
)

if TYPE_CHECKING:
    from srb.core.asset import (
        ManipulatorType,
        MobileManipulatorType,
        MobileRobotType,
        ObjectType,
        RobotType,
        SceneryType,
    )


def update_offline_srb_cache():
    if not SRB_CACHE_PATH.is_dir():
        SRB_CACHE_PATH.mkdir(parents=True)

    update_offline_srb_env_cache()
    update_offline_srb_robot_cache()
    update_offline_srb_scenery_cache()
    update_offline_srb_object_cache()


def update_offline_srb_env_cache():
    from srb.utils import logging
    from srb.utils.registry import get_srb_tasks

    if not is_isaacsim_initialized():
        logging.critical(
            "Updating the cache of registered environments will likely fail because Isaac Sim is not initialized"
        )

    from srb import tasks as _  # noqa: F401

    registered_envs = sorted(map(lambda env: env.removeprefix("srb/"), get_srb_tasks()))

    if not registered_envs:
        logging.warning(
            "Cannot update the cache of registered environments because no environments are registered"
        )
        return

    current_cache = read_offline_srb_env_cache()
    if registered_envs == current_cache:
        logging.trace("The cache of registered environments is up-to-date")
        return

    with SRB_ENV_CACHE_PATH.open("wb") as f:
        f.write(json.dumps(registered_envs))
    logging.debug(
        f"Updated the cache of registered environments to {SRB_ENV_CACHE_PATH}"
    )


def update_offline_srb_robot_cache():
    from srb.core.asset import (
        ManipulatorRegistry,
        ManipulatorType,
        MobileManipulatorRegistry,
        MobileManipulatorType,
        MobileRobotRegistry,
        MobileRobotType,
        RobotType,
    )
    from srb.utils import logging

    registry_map = {
        RobotType.MOBILE_ROBOT: (MobileRobotRegistry, MobileRobotType),
        RobotType.MANIPULATOR: (ManipulatorRegistry, ManipulatorType),
        RobotType.MOBILE_MANIPULATOR: (
            MobileManipulatorRegistry,
            MobileManipulatorType,
        ),
    }

    robots_cache = {}
    robots_cache_enum_keys = {}

    for robot_type, (registry, subtype_enum_cls) in registry_map.items():
        type_data = {
            subtype.name.lower(): sorted(robot.name() for robot in robots)
            for subtype, robots in registry.items()
            if robots
        }
        if type_data:
            robots_cache[robot_type.name.lower()] = type_data

            inner_enum_dict = {}
            for subtype_str, names in type_data.items():
                try:
                    subtype_enum = subtype_enum_cls[subtype_str.upper()]
                    inner_enum_dict[subtype_enum] = names
                except KeyError:
                    logging.warning(
                        f"Invalid robot subtype key '{subtype_str}' for type '{robot_type.name.lower()}', skipping."
                    )
            robots_cache_enum_keys[robot_type] = inner_enum_dict

    if not robots_cache:
        logging.warning(
            "Cannot update the cache of registered robots because no robots are registered"
        )
        return

    current_cache = read_offline_srb_robot_cache()
    if robots_cache_enum_keys == current_cache:
        logging.trace("The cache of registered robots is up-to-date")
        return

    with SRB_ROBOT_CACHE_PATH.open("wb") as f:
        f.write(json.dumps(robots_cache))
    logging.debug(f"Updated the cache of registered robots to {SRB_ROBOT_CACHE_PATH}")


def update_offline_srb_scenery_cache():
    from srb.core.asset import SceneryRegistry
    from srb.utils import logging

    if not is_isaacsim_initialized():
        logging.critical(
            "Updating the cache of registered sceneries will likely fail because Isaac Sim is not initialized"
        )

    from srb.assets import scenery as _  # noqa: F401

    sceneries = {}
    sceneries_enum_keys = {}

    for scenery_type, _sceneries in SceneryRegistry.items():
        if _sceneries:
            names = sorted(scenery.name() for scenery in _sceneries)
            sceneries[scenery_type.name.lower()] = names
            sceneries_enum_keys[scenery_type] = names

    if not sceneries:
        logging.warning(
            "Cannot update the cache of registered sceneries because no sceneries are registered"
        )
        return

    current_cache = read_offline_srb_scenery_cache()
    if sceneries_enum_keys == current_cache:
        logging.trace("The cache of registered sceneries is up-to-date")
        return

    with SRB_SCENERY_CACHE_PATH.open("wb") as f:
        f.write(json.dumps(sceneries))
    logging.debug(
        f"Updated the cache of registered sceneries to {SRB_SCENERY_CACHE_PATH}"
    )


def update_offline_srb_object_cache():
    from srb.core.asset import ObjectRegistry
    from srb.utils import logging

    if not is_isaacsim_initialized():
        logging.critical(
            "Updating the cache of registered objects will likely fail because Isaac Sim is not initialized"
        )

    from srb.assets import object as _  # noqa: F401

    objects = {}
    objects_enum_keys = {}

    for object_type, _objects in ObjectRegistry.items():
        if _objects:
            names = sorted(object.name() for object in _objects)
            objects[object_type.name.lower()] = names
            objects_enum_keys[object_type] = names

    if not objects:
        logging.warning(
            "Cannot update the cache of registered objects because no objects are registered"
        )
        return

    current_cache = read_offline_srb_object_cache()
    if objects_enum_keys == current_cache:
        logging.trace("The cache of registered objects is up-to-date")
        return

    with SRB_OBJECT_CACHE_PATH.open("wb") as f:
        f.write(json.dumps(objects))
    logging.debug(f"Updated the cache of registered objects to {SRB_OBJECT_CACHE_PATH}")


def read_offline_srb_env_cache() -> Sequence[str] | None:
    if not SRB_ENV_CACHE_PATH.exists():
        return None
    try:
        with SRB_ENV_CACHE_PATH.open("rb") as f:
            data = json.loads(f.read())
        if isinstance(data, list) and all(isinstance(item, str) for item in data):
            return data
        else:
            SRB_ENV_CACHE_PATH.unlink()
            return None
    except Exception:
        SRB_ENV_CACHE_PATH.unlink()
        return None


def read_offline_srb_robot_cache() -> (
    Dict[
        "RobotType",
        Dict[
            "MobileRobotType | ManipulatorType | MobileManipulatorType",
            Sequence[str],
        ],
    ]
    | None
):
    if not SRB_ROBOT_CACHE_PATH.exists():
        return None

    from srb.core.asset import (
        ManipulatorType,
        MobileManipulatorType,
        MobileRobotType,
        RobotType,
    )

    try:
        with SRB_ROBOT_CACHE_PATH.open("rb") as f:
            raw_cache = json.loads(f.read())
    except Exception:
        SRB_ROBOT_CACHE_PATH.unlink()
        return None

    if not isinstance(raw_cache, dict):
        SRB_ROBOT_CACHE_PATH.unlink()
        return None

    converted_cache: Dict[
        RobotType,
        Dict[
            Union[MobileRobotType, ManipulatorType, MobileManipulatorType],
            Sequence[str],
        ],
    ] = {}
    for rt_str, subtypes_dict in raw_cache.items():
        try:
            rt_enum = RobotType[rt_str.upper()]
            inner_converted_dict = {}
            subtype_enum_cls = None
            if rt_enum == RobotType.MOBILE_ROBOT:
                subtype_enum_cls = MobileRobotType
            elif rt_enum == RobotType.MANIPULATOR:
                subtype_enum_cls = ManipulatorType
            elif rt_enum == RobotType.MOBILE_MANIPULATOR:
                subtype_enum_cls = MobileManipulatorType

            if subtype_enum_cls and isinstance(subtypes_dict, dict):
                for subtype_str, names in subtypes_dict.items():
                    try:
                        subtype_enum = subtype_enum_cls[subtype_str.upper()]
                        inner_converted_dict[subtype_enum] = names
                    except Exception:
                        SRB_ROBOT_CACHE_PATH.unlink()
                        return None
            converted_cache[rt_enum] = inner_converted_dict
        except Exception:
            SRB_ROBOT_CACHE_PATH.unlink()
            return None

    return converted_cache


def read_offline_srb_scenery_cache() -> Dict["SceneryType", Sequence[str]] | None:
    if not SRB_SCENERY_CACHE_PATH.exists():
        return None

    from srb.core.asset import SceneryType

    try:
        with SRB_SCENERY_CACHE_PATH.open("rb") as f:
            raw_cache = json.loads(f.read())
    except Exception:
        SRB_SCENERY_CACHE_PATH.unlink()
        return None

    if not isinstance(raw_cache, dict):
        SRB_SCENERY_CACHE_PATH.unlink()
        return None

    converted_cache: Dict[SceneryType, Sequence[str]] = {}
    if isinstance(raw_cache, dict):
        for st_str, names in raw_cache.items():
            try:
                st_enum = SceneryType[st_str.upper()]
                if isinstance(names, list) and all(
                    isinstance(item, str) for item in names
                ):
                    converted_cache[st_enum] = names
                else:
                    SRB_SCENERY_CACHE_PATH.unlink()
                    return None
            except Exception:
                SRB_SCENERY_CACHE_PATH.unlink()
                return None
    else:
        SRB_SCENERY_CACHE_PATH.unlink()
        return None

    return converted_cache


def read_offline_srb_object_cache() -> Dict["ObjectType", Sequence[str]] | None:
    if not SRB_OBJECT_CACHE_PATH.exists():
        return None

    from srb.core.asset import ObjectType

    try:
        with SRB_OBJECT_CACHE_PATH.open("rb") as f:
            raw_cache = json.loads(f.read())
    except Exception:
        SRB_OBJECT_CACHE_PATH.unlink()
        return None

    if not isinstance(raw_cache, dict):
        SRB_OBJECT_CACHE_PATH.unlink()
        return None

    converted_cache: Dict[ObjectType, Sequence[str]] = {}
    if isinstance(raw_cache, dict):
        for ot_str, names in raw_cache.items():
            try:
                ot_enum = ObjectType[ot_str.upper()]
                if isinstance(names, list) and all(
                    isinstance(item, str) for item in names
                ):
                    converted_cache[ot_enum] = names
                else:
                    SRB_OBJECT_CACHE_PATH.unlink()
                    return None
            except Exception:
                SRB_OBJECT_CACHE_PATH.unlink()
                return None
    else:
        SRB_OBJECT_CACHE_PATH.unlink()
        return None

    return converted_cache



from .aerial_robot import AerialRobot, Multicopter  # noqa: F401
from .ground_robot import GroundRobot, LeggedRobot, WheeledRobot  # noqa: F401
from .mobile_robot import MobileRobot, MobileRobotRegistry  # noqa: F401
from .mobile_robot_type import MobileRobotType  # noqa: F401
from .orbital_robot import Lander, OrbitalRobot  # noqa: F401



from functools import cache
from typing import Tuple


@cache
def is_semver_compatible(
    required: Tuple[int, int, int], current: Tuple[int, int, int]
) -> bool:
    return current[0] == required[0] and (
        current[1] > required[1]
        or (current[1] == required[1] and current[2] >= required[2])
    )



from .facility import LunaLab, Oberpfaffenhofen  # noqa: F401
from .ground_plane import GroundPlane  # noqa: F401
from .planetary_surface import MarsSurface, MoonSurface  # noqa: F401
from .spacecraft import StaticGateway, StaticVenusExpress  # noqa: F401



import weakref
from collections.abc import Callable

import carb
import omni
from isaaclab.devices import Se3Keyboard as __Se3Keyboard


class KeyboardTeleopInterface(__Se3Keyboard):
    def __str__(self) -> str:
        msg = super().__str__()

        msg += "\n"
        msg += "\t----------------------------------------------\n"
        msg += "\tAdditional controls:\n"
        msg += "\tToggle gripper (alternative): R\n"
        return msg

    def _on_keyboard_event(self, event, *args, **kwargs) -> bool:
        ret = super()._on_keyboard_event(event, *args, **kwargs)

        if event.type == carb.input.KeyboardEventType.KEY_PRESS:  # type: ignore
            if event.input.name == "R":
                self._close_gripper = not self._close_gripper

        return ret


class EventKeyboardTeleopInterface:
    def __init__(self, callbacks: dict[str, Callable] = {}):
        self._appwindow = omni.appwindow.get_default_app_window()  # type: ignore
        self._input = carb.input.acquire_input_interface()  # type: ignore
        self._keyboard = self._appwindow.get_keyboard()
        self._keyboard_sub = self._input.subscribe_to_keyboard_events(
            self._keyboard,
            lambda event, *args, obj=weakref.proxy(self): obj._on_keyboard_event(
                event, *args
            ),
        )
        self._additional_callbacks = callbacks

    def __del__(self):
        self._input.unsubscribe_from_keyboard_events(self._keyboard, self._keyboard_sub)
        self._keyboard_sub = None

    def add_callback(self, key: str, func: Callable):
        self._additional_callbacks[key] = func

    def _on_keyboard_event(self, event, *args, **kwargs):
        if event.type == carb.input.KeyboardEventType.KEY_PRESS:  # type: ignore
            if event.input.name in self._additional_callbacks:
                self._additional_callbacks[event.input.name]()
        return True



from .manipulation import *  # noqa: F403
from .mobile import *  # noqa: F403
from .mobile_manipulation import *  # noqa: F403

# TODO[mid]: Check & fix transforms of frames on all robots (cameras, payload mount, manipulator mount)



from enum import Enum, auto
from typing import Type

from typing_extensions import Self


class InterfaceType(str, Enum):
    GUI = auto()
    ROS = auto()

    def __str__(self) -> str:
        return self.name.lower()

    @classmethod
    def from_str(cls, string: str) -> Self | None:
        return next(
            (variant for variant in cls if string.upper() == variant.name), None
        )

    def implementer(self) -> Type:
        match self:
            case InterfaceType.GUI:
                from .interface.gui import GuiInterface

                return GuiInterface
            case InterfaceType.ROS:
                from .interface.ros import RosInterface

                return RosInterface
            case _:
                raise NotImplementedError(f"Interface type {self} not implemented")


class TeleopDeviceType(str, Enum):
    KEYBOARD = auto()
    ROS = auto()
    GAMEPAD = auto()
    SPACEMOUSE = auto()
    HAPTIC = auto()

    def __str__(self) -> str:
        return self.name.lower()

    @classmethod
    def from_str(cls, string: str) -> Self | None:
        return next(
            (variant for variant in cls if string.upper() == variant.name), None
        )

    def implementer(self) -> Type:
        match self:
            case TeleopDeviceType.KEYBOARD:
                from .teleop.keyboard import KeyboardTeleopInterface

                return KeyboardTeleopInterface
            case TeleopDeviceType.ROS | TeleopDeviceType.GAMEPAD:
                from .teleop.ros import ROSTeleopInterface

                return ROSTeleopInterface
            case TeleopDeviceType.SPACEMOUSE:
                from .teleop.spacemouse import SpacemouseTeleopInterface

                return SpacemouseTeleopInterface
            case TeleopDeviceType.HAPTIC:
                from .teleop.haptic import HapticROSTeleopInterface

                return HapticROSTeleopInterface
            case _:
                raise NotImplementedError(f"Teleop device {self} not implemented")



from isaaclab.markers import *  # noqa: F403

from .arrow import ARROW_CFG  # noqa: F401
from .cone import CONE_CFG  # noqa: F401
from .cylinder import CYLINDER_CFG  # noqa: F401
from .frame import FRAME_MARKER_CFG, FRAME_MARKER_SMALL_CFG  # noqa: F401



from isaacsim.robot_setup.assembler import AssembledBodies as __AssembledBodies


class AssembledBodies(__AssembledBodies):
    def set_attach_path_root_joints_enabled(self, enabled: bool):
        for root_joint in self.root_joints:
            root_joint.GetProperty("physics:jointEnabled").Set(enabled)



from enum import Enum, auto
from typing import Tuple

from typing_extensions import Self


class Domain(str, Enum):
    ASTEROID = auto()
    EARTH = auto()
    MARS = auto()
    MOON = auto()
    ORBIT = auto()

    def __str__(self) -> str:
        return self.name.lower()

    @classmethod
    def from_str(cls, string: str) -> Self | None:
        return next(
            (variant for variant in cls if string.upper() == variant.name), None
        )

    @property
    def gravity_magnitude(self) -> float:
        """
        Magnitude of gravitational acceleration in m/s².

        - Asteroid: 50% gravitational acceleration of Ceres (largest body in the asteroid belt).
        - Orbit: No gravitational acceleration.
        """
        match self:
            case Domain.ASTEROID:
                return 0.14219
            case Domain.EARTH:
                return 9.80665
            case Domain.MARS:
                return 3.72076
            case Domain.MOON:
                return 1.62496
            case Domain.ORBIT:
                return 0.0

    @property
    def gravity_variation(self) -> float:
        """
        Difference between the maximum and minimum of gravitational acceleration in m/s².

        - Asteroid: Ceres is considered as the maximum (largest body in the asteroid belt).
        - Orbit: No gravitational acceleration.
        """
        match self:
            case Domain.ASTEROID:
                return 0.28438
            case Domain.EARTH:
                return 0.0698
            case Domain.MARS:
                return 0.0279
            case Domain.MOON:
                return 0.0253
            case Domain.ORBIT:
                return 0.0

    @property
    def gravity_range(self) -> Tuple[float, float]:
        """
        Range of gravitational acceleration in m/s² calculated as the magnitude ± variation/2.
        """
        magnitude = self.gravity_magnitude
        delta = self.gravity_variation / 2.0
        return (magnitude - delta, magnitude + delta)

    @property
    def light_intensity(self) -> float:
        """
        Intensity of Solar light in W/m².

        - Asteroid: Taken at 2.7 AU.
        - Earth | Mars: Taken at the surface. The peak value (sunny day) is subtracted by half of the variation.
        - Moon | Orbit: Taken at 1 AU.
        """
        match self:
            case Domain.ASTEROID:
                return 190.0
            case Domain.EARTH:
                return 775
            case Domain.MARS:
                return 729
            case Domain.MOON | Domain.ORBIT:
                return 1361.0

    @property
    def light_intensity_variation(self) -> float:
        """
        Difference between the maximum and minimum of Solar light intensity in W/m².

        - Asteroid: Approximate range between 2.55 and 2.97 AU.
        - Earth | Mars: Guesstimated effect of atmosphere and weather.
        - Moon | Orbit: Minor variation due to elliptical orbit.
        """
        match self:
            case Domain.ASTEROID:
                return 50.0
            case Domain.EARTH:
                return 450.0
            case Domain.MARS:
                return 226.0
            case Domain.MOON | Domain.ORBIT:
                return 0.5

    @property
    def light_intensity_range(self) -> Tuple[float, float]:
        """
        Range of Solar light intensity in W/m² calculated as the intensity ± variation/2.
        """
        intensity = self.light_intensity
        delta = self.light_intensity_variation / 2.0
        return (intensity - delta, intensity + delta)

    @property
    def light_angular_diameter(self) -> float:
        """
        Angular diameter of the Solar light source in degrees.

        - Earth | Mars: Taken at their distance from the Sun.
        - Asteroid | Moon | Orbit: Approximated as a point source due to lack of atmosphere.
        """
        match self:
            case Domain.EARTH:
                return 0.53
            case Domain.MARS:
                return 0.35
            case Domain.ASTEROID | Domain.MOON | Domain.ORBIT:
                return 0.0

    @property
    def light_angular_diameter_variation(self) -> float:
        """
        Variation of the angular diameter of the Solar light source in degrees.
        """
        match self:
            case Domain.EARTH:
                return 0.021
            case Domain.MARS:
                return 0.08
            case Domain.ASTEROID | Domain.MOON | Domain.ORBIT:
                return 0.0

    @property
    def light_angular_diameter_range(self) -> Tuple[float, float]:
        """
        Range of the angular diameter of the Solar light source in degrees calculated as the diameter ± variation/2.
        """
        diameter = self.light_angular_diameter
        delta = self.light_angular_diameter_variation / 2.0
        return (diameter - delta, diameter + delta)

    @property
    def light_color_temperature(self) -> float:
        """
        Temperature of the Solar light source in K.

        - Earth | Mars: Guesstimated effect of atmosphere and weather.
        - Asteroid | Moon | Orbit: Intrinsic color temperature of the Sun.
        """
        match self:
            case Domain.EARTH:
                return 5750.0
            case Domain.MARS:
                return 6250.0
            case Domain.ASTEROID | Domain.MOON | Domain.ORBIT:
                return 5778.0

    @property
    def light_color_temperature_variation(self) -> float:
        """
        Variation of the temperature of the Solar light source in K.

        - Earth | Mars: Guesstimated effect of atmosphere and weather.
        - Asteroid | Moon | Orbit: No significant variation.
        """
        match self:
            case Domain.EARTH:
                return 1500.0
            case Domain.MARS:
                return 500.0
            case Domain.ASTEROID | Domain.MOON | Domain.ORBIT:
                return 0.0

    @property
    def light_color_temperature_range(self) -> Tuple[float, float]:
        """
        Range of the temperature of the Solar light source in K calculated as the temperature ± variation/2.
        """

        temperature = self.light_color_temperature
        delta = self.light_color_temperature_variation / 2.0
        return (temperature - delta, temperature + delta)

    @property
    def contrastive_grayscale(self) -> float:
        """
        Grayscale value for contrastive color based on the domain.

        - Asteroid | Moon: Light gray.
        - Earth | Mars | Orbit: Dark gray.
        """
        match self:
            case Domain.ASTEROID | Domain.MOON:
                return 0.8
            case Domain.EARTH | Domain.MARS | Domain.ORBIT:
                return 0.1
            case _:
                return 0.7071



import threading
import time
from collections.abc import Callable
from typing import List

import numpy
import pyspacemouse
from scipy.spatial.transform import Rotation

from srb.interfaces.teleop import DeviceBase
from srb.utils import logging


# Driver: https://github.com/FreeSpacenav/spacenavd
class SpacemouseTeleopInterface(DeviceBase):
    def __init__(
        self,
        pos_sensitivity: float = 1.0,
        rot_sensitivity: float = 3.1415927,
        rate: float = 1000.0,
    ):
        self.pos_sensitivity = pos_sensitivity
        self.rot_sensitivity = rot_sensitivity
        self.sleep_rate = 1.0 / rate

        # Command buffers
        self._close_gripper = False
        self._delta_pos = numpy.zeros(3)  # (x, y, z)
        self._delta_rot = numpy.zeros(3)  # (roll, pitch, yaw)
        self._additional_callbacks = {}

        # Open the device
        try:
            success = pyspacemouse.open(
                dof_callback=self._cb_dof,  # type: ignore
                button_callback=self._cb_button,  # type: ignore
            )
            if success:
                # Run a background thread for the device
                self._thread = threading.Thread(target=self._run_device)
                self._thread.daemon = True
                self._thread.start()
            else:
                logging.error(
                    "Failed to open a SpaceMouse device. Is it connected?",
                )
        except Exception as e:
            logging.error(
                f"Failed to open a SpaceMouse device. Is it connected?\n{e}",
            )

    def __del__(self):
        if hasattr(self, "_thread"):
            self._thread.join()

    def __str__(self) -> str:
        msg = f"Spacemouse Controller ({self.__class__.__name__})\n"
        msg += "\tToggle gripper (alternative): Right button\n"
        msg += "\tReset: Left button\n"
        return msg

    def reset(self):
        self._close_gripper = False
        self._delta_pos = numpy.zeros(3)
        self._delta_rot = numpy.zeros(3)

    def add_callback(self, key: str, func: Callable):
        if key not in ["L", "R", "LR"]:
            raise ValueError(
                f"Only left (L), right (R), and right-left (LR) buttons supported. Provided: {key}."
            )
        self._additional_callbacks[key] = func

    def advance(self) -> tuple[numpy.ndarray, bool]:
        rot_vec = Rotation.from_euler("XYZ", self._delta_rot).as_rotvec()
        return numpy.concatenate([self._delta_pos, rot_vec]), self._close_gripper

    def _run_device(self):
        while True:
            _state = pyspacemouse.read()
            time.sleep(self.sleep_rate)

    def _cb_dof(self, state: pyspacemouse.SpaceNavigator):
        self._delta_pos = numpy.array(
            [
                state.y * self.pos_sensitivity,
                -state.x * self.pos_sensitivity,
                state.z * self.pos_sensitivity,
            ]
        )
        self._delta_rot = numpy.array(
            [
                -state.roll * self.rot_sensitivity,
                -state.pitch * self.rot_sensitivity,
                -state.yaw * self.rot_sensitivity,
            ]
        )

    def _cb_button(self, state: pyspacemouse.SpaceNavigator, buttons: List[bool]):
        if buttons[0]:
            self.reset()
            if "L" in self._additional_callbacks.keys():
                self._additional_callbacks["L"]()
        if buttons[1]:
            self._close_gripper = not self._close_gripper
            if "R" in self._additional_callbacks.keys():
                self._additional_callbacks["R"]()
        if all(buttons):
            if "LR" in self._additional_callbacks.keys():
                self._additional_callbacks["LR"]()



from __future__ import annotations

from functools import cached_property
from typing import ClassVar, Dict, Iterable, List, Sequence, Tuple, Type

from srb.core.action import ActionGroup
from srb.core.asset import ArticulationCfg, AssetRegistry, RigidObjectCfg
from srb.core.asset.asset import Asset
from srb.core.asset.asset_type import AssetType
from srb.core.asset.common import Frame
from srb.core.asset.robot.robot_type import RobotType


class Robot(Asset, asset_entrypoint=AssetType.ROBOT):
    ## Model
    asset_cfg: RigidObjectCfg | ArticulationCfg

    ## Actions
    actions: ActionGroup

    ## Frames
    frame_base: Frame

    def __init_subclass__(
        cls,
        robot_entrypoint: RobotType | None = None,
        robot_metaclass: bool = False,
        asset_metaclass: bool = False,
        **kwargs,
    ):
        super().__init_subclass__(
            asset_metaclass=(
                asset_metaclass or robot_entrypoint is not None or robot_metaclass
            ),
            **kwargs,
        )
        if robot_entrypoint is not None:
            assert isinstance(robot_entrypoint, RobotType), (
                f"Class '{cls.__name__}' is marked as a robot entrypoint, but '{robot_entrypoint}' is not a valid {RobotType}"
            )
            assert robot_entrypoint not in RobotRegistry.base_types.keys(), (
                f"Class '{cls.__name__}' is marked as '{robot_entrypoint}' robot entrypoint, but it was already marked by '{RobotRegistry.base_types[robot_entrypoint].__name__}'"
            )
            RobotRegistry.base_types[robot_entrypoint] = cls
        elif robot_metaclass:
            RobotRegistry.meta_types.append(cls)
        elif not asset_metaclass:
            # Note: Sorted here is a simple workaround to prioritize "mobile_manipulator" over "mobile_robot"
            for robot_type, base in sorted(RobotRegistry.base_types.items()):
                if issubclass(cls, base):
                    if robot_type not in RobotRegistry.registry.keys():
                        RobotRegistry.registry[robot_type] = []
                    else:
                        assert cls.name() not in (
                            robot.name() for robot in RobotRegistry.registry[robot_type]
                        ), (
                            f"Cannot register multiple robots with an identical name: '{cls.__module__}:{cls.__name__}' already exists as '{next(robot for robot in RobotRegistry.registry[robot_type] if cls.name() == robot.name()).__module__}:{cls.__name__}'"
                        )
                    RobotRegistry.registry[robot_type].append(cls)
                    break

    @cached_property
    def robot_type(self) -> RobotType:
        for robot_type, base in RobotRegistry.base_types.items():
            if isinstance(self, base):
                return robot_type
        raise ValueError(f"Class '{self.__class__.__name__}' has unknown robot type")

    @classmethod
    def robot_registry(cls) -> Sequence[Type[Robot]]:
        return list(RobotRegistry.values_inner())

    @classmethod
    def asset_registry(cls) -> Sequence[Type[Robot]]:
        return AssetRegistry.registry.get(AssetType.ROBOT, [])  # type: ignore


class RobotRegistry:
    registry: ClassVar[Dict[RobotType, List[Type[Robot]]]] = {}
    base_types: ClassVar[Dict[RobotType, Type[Robot]]] = {}
    meta_types: ClassVar[List[Type[Robot]]] = []

    @classmethod
    def keys(cls) -> Iterable[RobotType]:
        return cls.registry.keys()

    @classmethod
    def items(cls) -> Iterable[Tuple[RobotType, Sequence[Type[Robot]]]]:
        return cls.registry.items()

    @classmethod
    def values(cls) -> Iterable[Iterable[Type[Robot]]]:
        return cls.registry.values()

    @classmethod
    def values_inner(cls) -> Iterable[Type[Robot]]:
        return (robot for robots in cls.registry.values() for robot in robots)

    @classmethod
    def n_robots(cls) -> int:
        return sum(len(robots) for robots in cls.registry.values())

    @classmethod
    def registered_modules(cls) -> Iterable[str]:
        return {
            robot.__module__ for robots in cls.registry.values() for robot in robots
        }

    @classmethod
    def registered_packages(cls) -> Iterable[str]:
        return {module.split(".", maxsplit=1)[0] for module in cls.registered_modules()}

    @classmethod
    def get_by_name(cls, name: str) -> Type[Robot] | None:
        for robot in cls.values_inner():
            if robot.name() == name:
                return robot
        return None



import math
from itertools import count
from typing import TYPE_CHECKING, List, Sequence, Tuple

import isaacsim.core.utils.stage as stage_utils
import numpy
import omni.kit.commands
import omni.timeline
from isaaclab.sim import clone
from omni.physx.scripts import particleUtils as particle_utils
from omni.physx.scripts import physicsUtils as physics_utils
from pxr import Gf, Sdf, Usd, UsdGeom, Vt

from pxr import PhysxSchema  # isort: skip

if TYPE_CHECKING:
    from .cfg import GridParticlesSpawnerCfg, ParticlesSpawnerCfg


@clone
def spawn_particles_grid(
    prim_path: str,
    cfg: "GridParticlesSpawnerCfg",
    translation: Tuple[float, float, float] | None = None,
    orientation: Tuple[float, float, float, float] | None = None,
) -> Usd.Prim:
    # Ratio
    if cfg.ratio < 1.0:
        ratio_cbrt = math.pow(cfg.ratio, 1.0 / 3.0)
        dim_x, dim_y, dim_z = (
            math.floor(ratio_cbrt * cfg.dim_x),
            math.floor(ratio_cbrt * cfg.dim_y),
            math.floor(ratio_cbrt * cfg.dim_z),
        )
        particle_spacing = (
            cfg.particle_spacing + (1.0 / ratio_cbrt - 1.0) * cfg.particle_size
        )
    else:
        dim_x, dim_y, dim_z = cfg.dim_x, cfg.dim_y, cfg.dim_z
        particle_spacing = cfg.particle_spacing

    # Distribution
    positions = _create_particles_grid(
        lower=Gf.Vec3f(
            -(cfg.particle_size + particle_spacing) * dim_x / 2.0,
            -(cfg.particle_size + particle_spacing) * dim_y / 2.0,
            0.0,
        )
        + (
            Gf.Vec3f(
                translation[0],
                translation[1],
                translation[2],
            )
            if translation is not None
            else Gf.Vec3f(0.0, 0.0, 0.0)
        ),
        particle_spacing=cfg.particle_size + particle_spacing,
        dim_x=dim_x,
        dim_y=dim_y,
        dim_z=dim_z,
    )

    # Randomize velocities
    velocities = [
        Gf.Vec3f(
            numpy.random.uniform(cfg.velocity[0][0], cfg.velocity[0][1])
            if isinstance(cfg.velocity[0], Sequence)
            else cfg.velocity[0],
            numpy.random.uniform(cfg.velocity[1][0], cfg.velocity[1][1])
            if isinstance(cfg.velocity[1], Sequence)
            else cfg.velocity[1],
            numpy.random.uniform(cfg.velocity[2][0], cfg.velocity[2][1])
            if isinstance(cfg.velocity[2], Sequence)
            else cfg.velocity[2],
        )
        for _ in range(len(positions))
    ]

    return _spawn_particles(prim_path, cfg, positions, velocities)


@clone
def spawn_particles_pyramid(
    prim_path: str,
    cfg: "GridParticlesSpawnerCfg",
    translation: Tuple[float, float, float] | None = None,
    orientation: Tuple[float, float, float, float] | None = None,
) -> Usd.Prim:
    # Ratio
    if cfg.ratio < 1.0:
        ratio_cbrt = math.pow(cfg.ratio, 1.0 / 3.0)
        dim_x, dim_y, dim_z = (
            math.floor(ratio_cbrt * cfg.dim_x),
            math.floor(ratio_cbrt * cfg.dim_y),
            math.floor(ratio_cbrt * cfg.dim_z),
        )
        particle_spacing = (
            cfg.particle_spacing + (1.0 / ratio_cbrt - 1.0) * cfg.particle_size
        )
    else:
        dim_x, dim_y, dim_z = cfg.dim_x, cfg.dim_y, cfg.dim_z
        particle_spacing = cfg.particle_spacing

    # Distribution
    positions = []
    slope_x = math.floor(dim_x / dim_z)
    slope_y = math.floor(dim_y / dim_z)
    for z in range(dim_z):
        lower = Gf.Vec3f(
            -(cfg.particle_size + particle_spacing) * (dim_x - z * slope_x) / 2.0,
            -(cfg.particle_size + particle_spacing) * (dim_y - z * slope_y) / 2.0,
            z * (cfg.particle_size + particle_spacing),
        ) + (
            Gf.Vec3f(
                translation[0],
                translation[1],
                translation[2],
            )
            if translation is not None
            else Gf.Vec3f(0.0, 0.0, 0.0)
        )
        positions.extend(
            _create_particles_grid(
                lower=lower,
                particle_spacing=cfg.particle_size + particle_spacing,
                dim_x=max(1, dim_x - z * slope_x),
                dim_y=max(1, dim_y - z * slope_y),
                dim_z=1,
            )
        )

    # Randomize velocities
    velocities = [
        Gf.Vec3f(
            numpy.random.uniform(cfg.velocity[0][0], cfg.velocity[0][1])
            if isinstance(cfg.velocity[0], Sequence)
            else cfg.velocity[0],
            numpy.random.uniform(cfg.velocity[1][0], cfg.velocity[1][1])
            if isinstance(cfg.velocity[1], Sequence)
            else cfg.velocity[1],
            numpy.random.uniform(cfg.velocity[2][0], cfg.velocity[2][1])
            if isinstance(cfg.velocity[2], Sequence)
            else cfg.velocity[2],
        )
        for _ in range(len(positions))
    ]

    return _spawn_particles(prim_path, cfg, positions, velocities)


def _spawn_particles(
    prim_path: str,
    cfg: "ParticlesSpawnerCfg",
    positions: Sequence[Gf.Vec3f],
    velocities: Sequence[Gf.Vec3f],
) -> Usd.Prim:
    ## Extract stage
    stage = stage_utils.get_current_stage()

    ## Particle system
    # Find a unique name for the particle system (append digits to the name)
    for i in count(1):
        particle_system_path = Sdf.Path(
            f"{str(particle_utils.get_default_particle_system_path(stage))}{i}"
        )
        if not stage.GetPrimAtPath(particle_system_path):
            break
    # Determine the appropriate offsets based on the particle size
    particle_contact_offset = (
        cfg.particle_contact_offset
        if cfg.particle_contact_offset is not None
        else 0.8 * cfg.particle_size
    )
    # Create the particle system
    particle_system = particle_utils.add_physx_particle_system(
        stage=stage,
        particle_system_path=particle_system_path,
        particle_system_enabled=cfg.particle_system_enabled,
        contact_offset=(
            cfg.contact_offset
            if cfg.contact_offset is not None
            else particle_contact_offset
        ),
        rest_offset=(
            cfg.rest_offset
            if cfg.rest_offset is not None
            else 0.99 * particle_contact_offset
        ),
        particle_contact_offset=(
            cfg.particle_contact_offset
            if cfg.particle_contact_offset is not None
            else particle_contact_offset
        ),
        solid_rest_offset=(
            cfg.solid_rest_offset
            if cfg.solid_rest_offset is not None
            else 0.99 * 0.6 * particle_contact_offset
        ),
        fluid_rest_offset=(
            cfg.fluid_rest_offset
            if cfg.fluid_rest_offset is not None
            else 0.99 * 0.6 * particle_contact_offset
        ),
        enable_ccd=cfg.enable_ccd,
        solver_position_iterations=cfg.solver_position_iterations,
        max_depenetration_velocity=cfg.max_depenetration_velocity,
        wind=cfg.wind,
        max_neighborhood=cfg.max_neighborhood,
        neighborhood_scale=cfg.neighborhood_scale,
        max_velocity=cfg.max_velocity,
        global_self_collision_enabled=cfg.global_self_collision_enabled,
        non_particle_collision_enabled=cfg.non_particle_collision_enabled,
    )

    ## Material
    mtl_created_list = []
    omni.kit.commands.execute(
        "CreateAndBindMdlMaterialFromLibrary",
        mdl_name="OmniSurfacePresets.mdl",
        mtl_created_list=mtl_created_list,
    )
    mtl_path = mtl_created_list[0]
    omni.kit.commands.execute(
        "BindMaterial", prim_path=particle_system_path, material_path=mtl_path
    )
    particle_utils.add_pbd_particle_material(
        stage=stage,
        path=mtl_path,
        friction=cfg.friction,
        particle_friction_scale=cfg.particle_friction_scale,
        damping=cfg.damping,
        viscosity=cfg.viscosity,
        vorticity_confinement=cfg.vorticity_confinement,
        surface_tension=cfg.surface_tension,
        cohesion=cfg.cohesion,
        adhesion=cfg.adhesion,
        particle_adhesion_scale=cfg.particle_adhesion_scale,
        adhesion_offset_scale=cfg.adhesion_offset_scale,
        gravity_scale=cfg.gravity_scale,
        lift=cfg.lift,
        drag=cfg.drag,
        density=cfg.density,
        cfl_coefficient=cfg.cfl_coefficient,
    )
    physics_utils.add_physics_material_to_prim(
        stage, particle_system.GetPrim(), mtl_path
    )

    ## Visuals
    # Shadows
    if not cfg.cast_shadows:
        primvars_api = UsdGeom.PrimvarsAPI(particle_system)
        primvars_api.CreatePrimvar(  # type:ignore
            "doNotCastShadows", Sdf.ValueTypeNames.Bool
        ).Set(False)
    if cfg.fluid:
        # Particle anisotropy
        if cfg.particle_anisotropy_enabled:
            anisotropy_api = PhysxSchema.PhysxParticleAnisotropyAPI.Apply(
                particle_system.GetPrim()
            )
            anisotropy_api.CreateParticleAnisotropyEnabledAttr().Set(True)
            anisotropy_api.CreateScaleAttr().Set(cfg.anisotropy_scale)
            anisotropy_api.CreateMinAttr().Set(cfg.anisotropy_min)
            anisotropy_api.CreateMaxAttr().Set(cfg.anisotropy_max)

        # Particle smoothing
        if cfg.particle_smoothing_enabled:
            smoothing_api = PhysxSchema.PhysxParticleSmoothingAPI.Apply(
                particle_system.GetPrim()
            )
            smoothing_api.CreateParticleSmoothingEnabledAttr().Set(True)
            smoothing_api.CreateStrengthAttr().Set(cfg.smoothing_strength)

        # Particle isosurface
        if cfg.isosurface_enabled:
            isosurfaceAPI = PhysxSchema.PhysxParticleIsosurfaceAPI.Apply(
                particle_system.GetPrim()
            )
            isosurfaceAPI.CreateIsosurfaceEnabledAttr().Set(True)
            isosurfaceAPI.CreateMaxVerticesAttr().Set(cfg.isosurface_max_vertices)
            isosurfaceAPI.CreateMaxTrianglesAttr().Set(cfg.isosurface_max_triangles)
            isosurfaceAPI.CreateMaxSubgridsAttr().Set(cfg.isosurface_max_subgrids)
            if cfg.isosurface_grid_spacing is not None:
                isosurfaceAPI.CreateGridSpacingAttr().Set(cfg.isosurface_grid_spacing)
            if cfg.isosurface_surface_distance is not None:
                isosurfaceAPI.CreateSurfaceDistanceAttr().Set(
                    cfg.isosurface_surface_distance
                )
            isosurfaceAPI.CreateNumMeshSmoothingPassesAttr().Set(
                cfg.isosurface_mesh_smoothing_passes
            )
            isosurfaceAPI.CreateNumMeshNormalSmoothingPassesAttr().Set(
                cfg.isosurface_mesh_normal_smoothing_passes
            )
            isosurfaceAPI.CreateGridFilteringPassesAttr().Set(
                cfg.isosurface_grid_filtering_passes
            )
            if cfg.isosurface_grid_smoothing_radius is not None:
                isosurfaceAPI.CreateGridSmoothingRadiusAttr().Set(
                    cfg.isosurface_grid_smoothing_radius
                )

    ## Particle set
    prim = particle_utils.add_physx_particleset_points(
        stage=stage,
        path=Sdf.Path(prim_path),
        positions_list=Vt.Vec3fArray(positions),
        velocities_list=Vt.Vec3fArray(velocities),
        widths_list=[cfg.particle_size] * len(positions),
        particle_system_path=particle_system_path,
        self_collision=cfg.self_collision,
        fluid=cfg.fluid,
        particle_group=cfg.particle_group,
        particle_mass=0.0,
        density=cfg.density,
    )

    ## Visibility
    if not cfg.visible:
        visibility_attribute = prim.GetVisibilityAttr()
        visibility_attribute.Set("invisible")

    return prim


def _create_particles_grid(
    lower: Gf.Vec3f,
    particle_spacing: float,
    dim_x: int,
    dim_y: int,
    dim_z: int,
) -> List[Gf.Vec3f]:
    x = lower[0]
    y = lower[1]
    z = lower[2]
    positions = [Gf.Vec3f(0.0)] * dim_x * dim_y * dim_z
    index = 0
    for _ in range(dim_x):
        for _ in range(dim_y):
            for _ in range(dim_z):
                positions[index] = Gf.Vec3f(x, y, z)
                index += 1
                z = z + particle_spacing
            z = lower[2]
            y = y + particle_spacing
        y = lower[1]
        x = x + particle_spacing
    return positions



from srb.core.asset import DeformableObjectCfg, Object
from srb.core.sim import DeformableBodyPropertiesCfg, MassPropertiesCfg, UsdFileCfg
from srb.utils.path import SRB_ASSETS_DIR_SRB_OBJECT


class JugglingBall(Object):
    asset_cfg: DeformableObjectCfg = DeformableObjectCfg(
        prim_path="{ENV_REGEX_NS}/juggling_ball",
        spawn=UsdFileCfg(
            usd_path=(
                SRB_ASSETS_DIR_SRB_OBJECT.joinpath("juggling_ball.usdc").as_posix()
            ),
            mass_props=MassPropertiesCfg(density=1000.0),
            deformable_props=DeformableBodyPropertiesCfg(
                rest_offset=0.0, contact_offset=0.001
            ),
            # physics_material=DeformableBodyMaterialCfg(
            #     poissons_ratio=0.4, youngs_modulus=1e5
            # ),
        ),
        init_state=DeformableObjectCfg.InitialStateCfg(pos=(0.0, 0.0, 1.0)),
        debug_vis=True,
    )



from srb.assets.object.tool import SOArm100Gripper
from srb.core.action import (  # noqa: F401
    ActionGroup,
    DifferentialIKControllerCfg,
    DifferentialInverseKinematicsActionCfg,
    InverseKinematicsActionGroup,
    OperationalSpaceControlActionGroup,
    OperationalSpaceControllerActionCfg,
    OperationalSpaceControllerCfg,
)
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import ArticulationCfg, Frame, SerialManipulator, Tool, Transform
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import deg_to_rad, rpy_to_quat
from srb.utils.path import SRB_ASSETS_DIR_SRB_ROBOT


class SOArm100D5(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/so_arm100",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("manipulator")
            .joinpath("so_arm100_5dof.usdz")
            .as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=12,
                solver_velocity_iteration_count=1,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "Shoulder_Rotation": 0.0,
                "Shoulder_Pitch": deg_to_rad(-15.0),
                "Elbow": deg_to_rad(20.0),
                "Wrist_Pitch": deg_to_rad(5.0),
                "Wrist_Roll": 0.0,
            },
            pos=(0.0, 0.0, 0.015),
            rot=rpy_to_quat(0.0, 0.0, 90.0),
        ),
        actuators={
            "arm": ImplicitActuatorCfg(
                joint_names_expr=[".*"],
                velocity_limit=10.0,
                effort_limit=10.0,
                stiffness=100.0,
                damping=20.0,
            ),
        },
    )
    end_effector: Tool | None = SOArm100Gripper()

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=[".*"],
            base_name="Base",
            body_name="Fixed_Gripper",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="pinv",
            ),
            scale=0.05,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )
    # actions: ActionGroup = OperationalSpaceControlActionGroup(
    #     OperationalSpaceControllerActionCfg(
    #         asset_name="robot",
    #         joint_names=[".*"],
    #         body_name="Fixed_Gripper",
    #         controller_cfg=OperationalSpaceControllerCfg(
    #             target_types=["pose_rel"],
    #             impedance_mode="variable_kp",
    #             inertial_dynamics_decoupling=True,
    #             motion_stiffness_limits_task=(10.0, 250.0),
    #             motion_damping_ratio_task=1.0,
    #         ),
    #         position_scale=0.05,
    #         orientation_scale=0.05,
    #         stiffness_scale=120.0,
    #         body_offset=OperationalSpaceControllerActionCfg.OffsetCfg(),
    #     )
    # )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="Base")
    frame_flange: Frame = Frame(
        prim_relpath="Fixed_Gripper",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_base_camera: Frame = Frame(
        prim_relpath="Base/camera_base",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="Fixed_Gripper/camera_wrist",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )


class SOArm100D7(SerialManipulator):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/so_arm100",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("manipulator")
            .joinpath("so_arm100_7dof.usdz")
            .as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                disable_gravity=True,
                max_depenetration_velocity=5.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=False,
                solver_position_iteration_count=12,
                solver_velocity_iteration_count=1,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            joint_pos={
                "Shoulder_Pitch": deg_to_rad(120.0),
                "Shoulder_Yaw": 0.0,
                "Humeral_Rotation": 0.0,
                "Elbow": deg_to_rad(-90.0),
                "Wrist_Roll": 0.0,
                "Wrist_Yaw": 0.0,
                "Wrist_Pitch": deg_to_rad(-90.0),
            },
            rot=rpy_to_quat(180.0, 0.0, 0.0),
        ),
        actuators={
            "arm": ImplicitActuatorCfg(
                joint_names_expr=[".*"],
                velocity_limit=10.0,
                effort_limit=10.0,
                stiffness=100.0,
                damping=20.0,
            ),
        },
    )

    ## Actions
    actions: ActionGroup = InverseKinematicsActionGroup(
        DifferentialInverseKinematicsActionCfg(
            asset_name="robot",
            joint_names=[".*"],
            base_name="Base",
            body_name="End_Servo",
            controller=DifferentialIKControllerCfg(
                command_type="pose",
                use_relative_mode=True,
                ik_method="svd",
            ),
            scale=0.05,
            body_offset=DifferentialInverseKinematicsActionCfg.OffsetCfg(),
        ),
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="Base")
    frame_flange: Frame = Frame(
        prim_relpath="End_Servo",
        offset=Transform(
            pos=(0.0, 0.0, -0.05),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_base_camera: Frame = Frame(
        prim_relpath="Base/camera_base",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_wrist_camera: Frame = Frame(
        prim_relpath="End_Servo/camera_wrist",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )



from ._metaclass import CombinedMobileManipulator  # noqa: F401
from .aerial_manipulator import AerialManipulator  # noqa: F401
from .ground_manipulator import GroundManipulator  # noqa: F401
from .orbital_manipulator import OrbitalManipulator  # noqa: F401



from typing import TYPE_CHECKING, Dict, List, Sequence, Tuple

import torch
from pxr import Gf

from srb.core.asset import Articulation, RigidObject, RigidObjectCollection, XFormPrim
from srb.core.manager import SceneEntityCfg
from srb.utils.math import quat_from_euler_xyz, quat_mul
from srb.utils.sampling import (
    sample_poisson_disk_2d_looped,
    sample_poisson_disk_3d_looped,
    sample_uniform,
)
from srb.utils.usd import safe_set_attribute_on_usd_prim

if TYPE_CHECKING:
    from srb._typing import AnyEnv


def reset_scene_to_default(env: "AnyEnv", env_ids: torch.Tensor):
    reset_rigid_objects_default(env, env_ids)
    reset_articulations_default(env, env_ids)
    reset_deformable_objects_default(env, env_ids)


def reset_rigid_objects_default(env: "AnyEnv", env_ids: torch.Tensor | None):
    for rigid_object in env.scene.rigid_objects.values():
        # Obtain default and deal with the offset for env origins
        default_root_state = rigid_object.data.default_root_state[env_ids].clone()
        default_root_state[:, 0:3] += env.scene.env_origins[env_ids]
        # Set into the physics simulation
        rigid_object.write_root_pose_to_sim(
            default_root_state[:, :7],
            env_ids=env_ids,  # type: ignore
        )
        # TODO[mid]: Do not reset velocity for kinematic objects
        rigid_object.write_root_velocity_to_sim(
            default_root_state[:, 7:],
            env_ids=env_ids,  # type: ignore
        )


def reset_articulations_default(env: "AnyEnv", env_ids: torch.Tensor | None):
    for articulation_asset in env.scene.articulations.values():
        # Obtain default and deal with the offset for env origins
        default_root_state = articulation_asset.data.default_root_state[env_ids].clone()
        default_root_state[:, 0:3] += env.scene.env_origins[env_ids]
        # Set into the physics simulation
        articulation_asset.write_root_pose_to_sim(
            default_root_state[:, :7],
            env_ids=env_ids,  # type: ignore
        )
        articulation_asset.write_root_velocity_to_sim(
            default_root_state[:, 7:],
            env_ids=env_ids,  # type: ignore
        )
        # Obtain default joint positions
        default_joint_pos = articulation_asset.data.default_joint_pos[env_ids].clone()
        default_joint_vel = articulation_asset.data.default_joint_vel[env_ids].clone()
        # Set into the physics simulation
        articulation_asset.write_joint_state_to_sim(
            default_joint_pos,
            default_joint_vel,
            env_ids=env_ids,  # type: ignore
        )


def reset_deformable_objects_default(env: "AnyEnv", env_ids: torch.Tensor | None):
    for deformable_object in env.scene.deformable_objects.values():
        # Obtain default and set into the physics simulation
        nodal_state = deformable_object.data.default_nodal_state_w[env_ids].clone()
        deformable_object.write_nodal_state_to_sim(nodal_state, env_ids=env_ids)  # type: ignore


def randomize_pose(
    env: "AnyEnv",
    env_ids: torch.Tensor | None,
    env_attr_name: str,
    pose_range: Dict[str, Tuple[float, float]],
):
    _env: "AnyEnv" = env.unwrapped  # type: ignore
    if env_ids is None:
        env_ids = torch.arange(
            _env.num_envs,
            device=_env.device,
        )

    range_list = [
        pose_range.get(key, (0.0, 0.0))
        for key in ["x", "y", "z", "roll", "pitch", "yaw"]
    ]
    ranges = torch.tensor(range_list, dtype=torch.float32, device=_env.device)
    rand_samples = sample_uniform(
        ranges[:, 0],
        ranges[:, 1],
        (len(env_ids), 6),
        device=_env.device,
    )
    positions = env.scene.env_origins[env_ids] + rand_samples[:, 0:3]
    orientations = quat_from_euler_xyz(
        rand_samples[:, 3], rand_samples[:, 4], rand_samples[:, 5]
    )

    pose_attr = getattr(env, env_attr_name)
    pose_attr[env_ids] = torch.cat([positions, orientations], dim=-1)


def randomize_pos(
    env: "AnyEnv",
    env_ids: torch.Tensor | None,
    env_attr_name: str,
    pos_range: Dict[str, Tuple[float, float]],
):
    _env: "AnyEnv" = env.unwrapped  # type: ignore
    if env_ids is None:
        env_ids = torch.arange(
            _env.num_envs,
            device=_env.device,
        )

    range_list = [pos_range.get(key, (0.0, 0.0)) for key in ["x", "y", "z"]]
    ranges = torch.tensor(range_list, dtype=torch.float32, device=_env.device)
    rand_samples = sample_uniform(
        ranges[:, 0],
        ranges[:, 1],
        (len(env_ids), 3),
        device=_env.device,
    )
    positions = env.scene.env_origins[env_ids] + rand_samples[:, 0:3]

    pos_attr = getattr(env, env_attr_name)
    pos_attr[env_ids] = positions


# Global dictionary to store velocity vectors for natural movement
_natural_movement_velocities = {}


def offset_pos_natural(
    env: "AnyEnv",
    env_ids: torch.Tensor | None,
    env_attr_name: str,
    axes: Sequence[str],
    step_range: Tuple[float, float],
    smoothness: float,
    pos_bounds: Dict[str, Tuple[float, float]],
):
    """Move the target position naturally with smoothed random changes in direction.

    Args:
        env: Environment instance
        env_ids: Indices of environments to update
        env_attr_name: Name of the attribute to modify
        axes: Which axes to apply movement to (e.g., ["x", "y"])
        step_range: Range of step sizes per update
        smoothness: Value between 0-1 controlling continuity of movement (higher = smoother)
        pos_bounds: Dictionary of position bounds for each axis
    """
    _env: "AnyEnv" = env.unwrapped  # type: ignore
    if env_ids is None:
        env_ids = torch.arange(
            _env.num_envs,
            device=_env.device,
        )

    # Create a state key for this specific environment and attribute
    state_key = f"{id(env)}:{env_attr_name}"

    # Get or initialize velocity vectors
    if state_key not in _natural_movement_velocities:
        # Initialize with random velocities
        _natural_movement_velocities[state_key] = torch.zeros(
            (_env.num_envs, 3), device=_env.device
        )
        # Set initial random direction for each environment
        for i in range(_env.num_envs):
            _natural_movement_velocities[state_key][i] = torch.randn(
                3, device=_env.device
            )
            # Normalize to ensure we start with unit vector
            _natural_movement_velocities[state_key][i] /= (
                torch.norm(_natural_movement_velocities[state_key][i]) + 1e-8
            )

    # Get the velocities for the current environments
    velocities = _natural_movement_velocities[state_key][env_ids]

    # Apply random perturbation with smoothing
    random_direction = torch.randn_like(velocities)
    # Normalize random direction
    random_direction /= torch.norm(random_direction, dim=1, keepdim=True) + 1e-8

    # Update velocity with smoothed random changes
    velocities = smoothness * velocities + (1 - smoothness) * random_direction
    # Normalize velocities to prevent acceleration
    velocities /= torch.norm(velocities, dim=1, keepdim=True) + 1e-8

    # Sample step sizes from step_range
    step_sizes = sample_uniform(
        step_range[0], step_range[1], (len(env_ids),), device=_env.device
    )

    # Apply step sizes to velocities
    delta_pos = velocities * step_sizes.unsqueeze(1)

    # Get current positions
    pos_attr = getattr(env, env_attr_name)
    current_positions = pos_attr[env_ids].clone()

    # Apply movement only to specified axes
    axis_indices = {"x": 0, "y": 1, "z": 2}
    active_axes = [axis_indices[axis] for axis in axes if axis in axis_indices]

    # Calculate new positions
    new_positions = current_positions.clone()
    for axis_idx in active_axes:
        new_positions[:, axis_idx] += delta_pos[:, axis_idx]

    # Handle boundaries by reflecting velocities when hitting bounds
    for axis in axes:
        if axis in pos_bounds:
            axis_idx = axis_indices[axis]
            min_bound, max_bound = pos_bounds[axis]

            # Calculate bounds relative to environment origins
            actual_min = min_bound + env.scene.env_origins[env_ids, axis_idx]
            actual_max = max_bound + env.scene.env_origins[env_ids, axis_idx]

            # Check if any positions exceed bounds
            below_min = new_positions[:, axis_idx] < actual_min
            above_max = new_positions[:, axis_idx] > actual_max

            # Reflect positions and velocities at boundaries
            if below_min.any():
                new_positions[below_min, axis_idx] = (
                    2 * actual_min[below_min] - new_positions[below_min, axis_idx]
                )
                velocities[below_min, axis_idx] *= -1.0

            if above_max.any():
                new_positions[above_max, axis_idx] = (
                    2 * actual_max[above_max] - new_positions[above_max, axis_idx]
                )
                velocities[above_max, axis_idx] *= -1.0

    # Save updated velocities back to the state dictionary
    _natural_movement_velocities[state_key][env_ids] = velocities

    # Update the positions in the environment
    pos_attr[env_ids] = new_positions


def randomize_command(
    env: "AnyEnv",
    env_ids: torch.Tensor | None,
    env_attr_name: str,
    magnitude: float = 1.0,
):
    _env: "AnyEnv" = env.unwrapped  # type: ignore
    if env_ids is None:
        env_ids = torch.arange(
            _env.num_envs,
            device=_env.device,
        )
    cmd_attr = getattr(env, env_attr_name)
    cmd_attr[env_ids] = sample_uniform(
        -magnitude,
        magnitude,
        (len(env_ids), *cmd_attr.shape[1:]),
        device=_env.device,
    )


def release_assembly_root_joins_on_action(
    env: "AnyEnv",
    env_ids: torch.Tensor | None,
    assembly_key: str,
    env_joint_assemblies_attr_name: str = "joint_assemblies",
    env_action_manager_attr_name: str = "action_manager",
    action_idx: int = 0,
    cmp_op: str = ">",
    cmp_value: float = 0.0,
):
    if env_ids is None:
        _env: "AnyEnv" = env.unwrapped  # type: ignore
        env_ids = torch.arange(
            _env.num_envs,
            device=_env.device,
        )

    joint_assembly = getattr(env, env_joint_assemblies_attr_name)[assembly_key]
    actions = getattr(env, env_action_manager_attr_name).action[env_ids, action_idx]

    for assembly, action in zip(joint_assembly, actions):
        assembly.set_attach_path_root_joints_enabled(
            eval(f"{action}{cmp_op}{cmp_value}")
        )


def reset_xform_orientation_uniform(
    env: "AnyEnv",
    env_ids: torch.Tensor,
    orientation_distribution_params: Dict[str, Tuple[float, float]],
    asset_cfg: SceneEntityCfg,
):
    asset: XFormPrim = env.scene[asset_cfg.name]

    range_list = [
        orientation_distribution_params.get(key, (0.0, 0.0))
        for key in ["roll", "pitch", "yaw"]
    ]
    ranges = torch.tensor(range_list, device=asset._device)
    rand_samples = sample_uniform(
        ranges[:, 0], ranges[:, 1], (1, 3), device=asset._device
    )

    orientations = quat_from_euler_xyz(
        rand_samples[:, 0], rand_samples[:, 1], rand_samples[:, 2]
    )

    asset.set_world_poses(orientations=orientations)


def randomize_usd_prim_attribute_uniform(
    env: "AnyEnv",
    env_ids: torch.Tensor | None,
    attr_name: str,
    distribution_params: Tuple[float | Sequence[float], float | Sequence[float]],
    asset_cfg: SceneEntityCfg,
):
    asset: XFormPrim = env.scene[asset_cfg.name]
    if isinstance(distribution_params[0], Sequence):
        dist_len = len(distribution_params[0])
        distribution_params = (  # type: ignore
            torch.tensor(distribution_params[0]),
            torch.tensor(distribution_params[1]),
        )
    else:
        dist_len = 1
    for i, prim in enumerate(asset.prims):
        if env_ids is not None and i not in env_ids:
            continue
        value = sample_uniform(
            distribution_params[0],  # type: ignore
            distribution_params[1],  # type: ignore
            (dist_len,),
            device="cpu",
        )
        value = value.item() if dist_len == 1 else value.tolist()
        safe_set_attribute_on_usd_prim(
            prim, f"inputs:{attr_name}", value, camel_case=True
        )


def randomize_gravity_uniform(
    env: "AnyEnv",
    env_ids: torch.Tensor | None,
    distribution_params: Tuple[Tuple[float, float, float], Tuple[float, float, float]],
):
    physics_scene = env.sim._physics_context._physics_scene  # type: ignore
    gravity = sample_uniform(
        torch.tensor(distribution_params[0]),
        torch.tensor(distribution_params[1]),
        (3,),
        device="cpu",
    )
    gravity_magnitude = torch.norm(gravity)
    if gravity_magnitude == 0.0:
        gravity_direction = gravity
    else:
        gravity_direction = gravity / gravity_magnitude

    physics_scene.CreateGravityDirectionAttr(Gf.Vec3f(*gravity_direction.tolist()))
    physics_scene.CreateGravityMagnitudeAttr(gravity_magnitude.item())


def follow_xform_orientation_linear_trajectory(
    env: "AnyEnv",
    env_ids: torch.Tensor,
    orientation_step_params: Dict[str, float],
    asset_cfg: SceneEntityCfg,
):
    asset: XFormPrim = env.scene[asset_cfg.name]

    _, current_quat = asset.get_world_poses()

    steps = torch.tensor(
        [orientation_step_params.get(key, 0.0) for key in ["roll", "pitch", "yaw"]],
        device=asset._device,
    )
    step_quat = quat_from_euler_xyz(steps[0], steps[1], steps[2]).unsqueeze(0)

    orientations = quat_mul(current_quat, step_quat)  # type: ignore

    asset.set_world_poses(orientations=orientations)


def reset_root_state_uniform_poisson_disk_2d(
    env: "AnyEnv",
    env_ids: torch.Tensor,
    pose_range: dict[str, tuple[float, float]],
    velocity_range: dict[str, tuple[float, float]],
    radius: float,
    asset_cfg: List[SceneEntityCfg],
):
    # Extract the used quantities (to enable type-hinting)
    assets: List[RigidObject | Articulation] = [
        env.scene[cfg.name] for cfg in asset_cfg
    ]
    # Get default root state
    root_states = torch.stack(
        [asset.data.default_root_state[env_ids].clone() for asset in assets],
    ).swapaxes(0, 1)

    # Poses
    range_list = [
        pose_range.get(key, (0.0, 0.0))
        for key in ["x", "y", "z", "roll", "pitch", "yaw"]
    ]
    ranges = torch.tensor(range_list, dtype=torch.float32, device=assets[0].device)
    samples_pos_xy = torch.tensor(
        sample_poisson_disk_2d_looped(
            (len(env_ids), len(asset_cfg)),
            (
                (range_list[0][0], range_list[1][0]),
                (range_list[0][1], range_list[1][1]),
            ),
            radius,
        ),
        device=assets[0].device,
    )
    rand_samples = sample_uniform(
        ranges[2:, 0],
        ranges[2:, 1],
        (len(env_ids), len(asset_cfg), 4),
        device=assets[0].device,
    )
    rand_samples = torch.cat([samples_pos_xy, rand_samples], dim=-1)

    positions = (
        root_states[:, :, 0:3]
        + env.scene.env_origins[env_ids].repeat(len(asset_cfg), 1, 1).swapaxes(0, 1)
        + rand_samples[:, :, 0:3]
    )
    orientations_delta = quat_from_euler_xyz(
        rand_samples[:, :, 3], rand_samples[:, :, 4], rand_samples[:, :, 5]
    )
    orientations = quat_mul(root_states[:, :, 3:7], orientations_delta)

    # Velocities
    range_list = [
        velocity_range.get(key, (0.0, 0.0))
        for key in ["x", "y", "z", "roll", "pitch", "yaw"]
    ]
    ranges = torch.tensor(range_list, dtype=torch.float32, device=assets[0].device)
    rand_samples = sample_uniform(
        ranges[:, 0],
        ranges[:, 1],
        (len(env_ids), len(asset_cfg), 6),
        device=assets[0].device,
    )
    velocities = root_states[:, :, 7:13] + rand_samples

    # Set into the physics simulation
    for asset, position, orientation, velocity in zip(
        assets,
        positions.unbind(1),
        orientations.unbind(1),
        velocities.unbind(1),
    ):
        asset.write_root_pose_to_sim(
            torch.cat([position, orientation], dim=-1),
            env_ids=env_ids,  # type: ignore
        )
        asset.write_root_velocity_to_sim(velocity, env_ids=env_ids)  # type: ignore


def reset_collection_root_state_uniform_poisson_disk_2d(
    env: "AnyEnv",
    env_ids: torch.Tensor,
    pose_range: dict[str, tuple[float, float]],
    velocity_range: dict[str, tuple[float, float]],
    radius: float,
    asset_cfg: SceneEntityCfg,
):
    # Extract the used quantities (to enable type-hinting)
    assets: RigidObjectCollection = env.scene[asset_cfg.name]

    # Get default root state
    root_states = assets.data.default_object_state[env_ids]

    # Poses
    range_list = [
        pose_range.get(key, (0.0, 0.0))
        for key in ["x", "y", "z", "roll", "pitch", "yaw"]
    ]
    ranges = torch.tensor(range_list, dtype=torch.float32, device=assets.device)
    samples_pos_xy = torch.tensor(
        sample_poisson_disk_2d_looped(
            (len(env_ids), assets.num_objects),
            (
                (range_list[0][0], range_list[1][0]),
                (range_list[0][1], range_list[1][1]),
            ),
            radius,
        ),
        device=assets.device,
    )
    rand_samples = sample_uniform(
        ranges[2:, 0],
        ranges[2:, 1],
        (len(env_ids), assets.num_objects, 4),
        device=assets.device,
    )
    rand_samples = torch.cat([samples_pos_xy, rand_samples], dim=-1)

    positions = (
        root_states[:, :, 0:3]
        + env.scene.env_origins[env_ids].repeat(assets.num_objects, 1, 1).swapaxes(0, 1)
        + rand_samples[:, :, 0:3]
    )
    orientations_delta = quat_from_euler_xyz(
        rand_samples[:, :, 3], rand_samples[:, :, 4], rand_samples[:, :, 5]
    )
    orientations = quat_mul(root_states[:, :, 3:7], orientations_delta)

    # Velocities
    range_list = [
        velocity_range.get(key, (0.0, 0.0))
        for key in ["x", "y", "z", "roll", "pitch", "yaw"]
    ]
    ranges = torch.tensor(range_list, dtype=torch.float32, device=assets.device)
    rand_samples = sample_uniform(
        ranges[:, 0],
        ranges[:, 1],
        (len(env_ids), assets.num_objects, 6),
        device=assets.device,
    )
    velocities = root_states[:, :, 7:13] + rand_samples

    # Set into the physics simulation
    assets.write_object_pose_to_sim(
        torch.cat([positions, orientations], dim=-1), env_ids=env_ids
    )
    assets.write_object_velocity_to_sim(velocities, env_ids=env_ids)


def reset_root_state_uniform_poisson_disk_3d(
    env: "AnyEnv",
    env_ids: torch.Tensor,
    pose_range: dict[str, tuple[float, float, float]],
    velocity_range: dict[str, tuple[float, float, float]],
    radius: float,
    asset_cfg: List[SceneEntityCfg],
):
    # Extract the used quantities (to enable type-hinting)
    assets: List[RigidObject | Articulation] = [
        env.scene[cfg.name] for cfg in asset_cfg
    ]
    # Get default root state
    root_states = torch.stack(
        [asset.data.default_root_state[env_ids].clone() for asset in assets],
    ).swapaxes(0, 1)

    # Poses
    range_list = [
        pose_range.get(key, (0.0, 0.0))
        for key in ["x", "y", "z", "roll", "pitch", "yaw"]
    ]
    ranges = torch.tensor(range_list, dtype=torch.float32, device=assets[0].device)
    samples_pos = torch.tensor(
        sample_poisson_disk_3d_looped(
            (len(env_ids), len(asset_cfg)),
            (
                (range_list[0][0], range_list[1][0], range_list[2][0]),
                (range_list[0][1], range_list[1][1], range_list[2][1]),
            ),
            radius,
        ),
        device=assets[0].device,
    )
    rand_samples = sample_uniform(
        ranges[3:, 0],
        ranges[3:, 1],
        (len(env_ids), len(asset_cfg), 3),
        device=assets[0].device,
    )
    rand_samples = torch.cat([samples_pos, rand_samples], dim=-1)

    positions = (
        root_states[:, :, 0:3]
        + env.scene.env_origins[env_ids].repeat(len(asset_cfg), 1, 1).swapaxes(0, 1)
        + rand_samples[:, :, 0:3]
    )
    orientations_delta = quat_from_euler_xyz(
        rand_samples[:, :, 3], rand_samples[:, :, 4], rand_samples[:, :, 5]
    )
    orientations = quat_mul(root_states[:, :, 3:7], orientations_delta)

    # Velocities
    range_list = [
        velocity_range.get(key, (0.0, 0.0))
        for key in ["x", "y", "z", "roll", "pitch", "yaw"]
    ]
    ranges = torch.tensor(range_list, dtype=torch.float32, device=assets[0].device)
    rand_samples = sample_uniform(
        ranges[:, 0],
        ranges[:, 1],
        (len(env_ids), len(asset_cfg), 6),
        device=assets[0].device,
    )
    velocities = root_states[:, :, 7:13] + rand_samples

    # Set into the physics simulation
    for asset, position, orientation, velocity in zip(
        assets,
        positions.unbind(1),
        orientations.unbind(1),
        velocities.unbind(1),
    ):
        asset.write_root_pose_to_sim(
            torch.cat([position, orientation], dim=-1),
            env_ids=env_ids,  # type: ignore
        )
        asset.write_root_velocity_to_sim(velocity, env_ids=env_ids)  # type: ignore


def reset_collection_root_state_uniform_poisson_disk_3d(
    env: "AnyEnv",
    env_ids: torch.Tensor,
    pose_range: dict[str, tuple[float, float, float]],
    velocity_range: dict[str, tuple[float, float, float]],
    radius: float,
    asset_cfg: SceneEntityCfg,
):
    # Extract the used quantities (to enable type-hinting)
    assets: RigidObjectCollection = env.scene[asset_cfg.name]

    # Get default root state
    root_states = assets.data.default_object_state[env_ids]

    # Poses
    range_list = [
        pose_range.get(key, (0.0, 0.0))
        for key in ["x", "y", "z", "roll", "pitch", "yaw"]
    ]
    ranges = torch.tensor(range_list, dtype=torch.float32, device=assets.device)
    samples_pos = torch.tensor(
        sample_poisson_disk_3d_looped(
            (len(env_ids), assets.num_objects),
            (
                (range_list[0][0], range_list[1][0], range_list[2][0]),
                (range_list[0][1], range_list[1][1], range_list[2][1]),
            ),
            radius,
        ),
        device=assets.device,
    )
    rand_samples = sample_uniform(
        ranges[3:, 0],
        ranges[3:, 1],
        (len(env_ids), assets.num_objects, 3),
        device=assets.device,
    )
    rand_samples = torch.cat([samples_pos, rand_samples], dim=-1)

    positions = (
        root_states[:, :, 0:3]
        + env.scene.env_origins[env_ids].repeat(assets.num_objects, 1, 1).swapaxes(0, 1)
        + rand_samples[:, :, 0:3]
    )
    orientations_delta = quat_from_euler_xyz(
        rand_samples[:, :, 3], rand_samples[:, :, 4], rand_samples[:, :, 5]
    )
    orientations = quat_mul(root_states[:, :, 3:7], orientations_delta)

    # Velocities
    range_list = [
        velocity_range.get(key, (0.0, 0.0))
        for key in ["x", "y", "z", "roll", "pitch", "yaw"]
    ]
    ranges = torch.tensor(range_list, dtype=torch.float32, device=assets.device)
    rand_samples = sample_uniform(
        ranges[:, 0],
        ranges[:, 1],
        (len(env_ids), assets.num_objects, 6),
        device=assets.device,
    )
    velocities = root_states[:, :, 7:13] + rand_samples

    # Set into the physics simulation
    assets.write_object_pose_to_sim(
        torch.cat([positions, orientations], dim=-1), env_ids=env_ids
    )
    assets.write_object_velocity_to_sim(velocities, env_ids=env_ids)



from dataclasses import MISSING

from srb.core.asset import AssetVariant, Humanoid, MobileRobot
from srb.core.env import BaseEventCfg, BaseSceneCfg, DirectEnv, DirectEnvCfg
from srb.core.marker import RED_ARROW_X_MARKER_CFG
from srb.core.sensor import Imu, ImuCfg
from srb.utils.cfg import configclass


@configclass
class MobileSceneCfg(BaseSceneCfg):
    imu_robot: ImuCfg = ImuCfg(
        prim_path=MISSING,  # type: ignore
        gravity_bias=(0.0, 0.0, 0.0),
        visualizer_cfg=RED_ARROW_X_MARKER_CFG.replace(  # type: ignore
            prim_path="/Visuals/imu_robot/lin_acc"
        ),
    )


@configclass
class MobileEventCfg(BaseEventCfg):
    pass


@configclass
class MobileEnvCfg(DirectEnvCfg):
    ## Assets
    robot: MobileRobot | Humanoid | AssetVariant = MISSING  # type: ignore
    _robot: MobileRobot | Humanoid = MISSING  # type: ignore

    ## Scene
    scene: MobileSceneCfg = MobileSceneCfg()

    ## Events
    events: MobileEventCfg = MobileEventCfg()

    def __post_init__(self):
        super().__post_init__()

        # Sensor: Robot IMU
        if self._robot.frame_imu:
            self.scene.imu_robot.prim_path = (
                f"{self.scene.robot.prim_path}/{self._robot.frame_base.prim_relpath}"
            )
            self.scene.imu_robot.offset.pos = self._robot.frame_imu.offset.pos
            self.scene.imu_robot.offset.rot = self._robot.frame_imu.offset.rot
        else:
            self.scene.imu_robot.prim_path = (
                f"{self.scene.robot.prim_path}/{self._robot.frame_base.prim_relpath}"
                if self._robot.frame_base.prim_relpath
                else self.scene.robot.prim_path
            )
            self.scene.imu_robot.offset.pos = self._robot.frame_base.offset.pos
            self.scene.imu_robot.offset.rot = self._robot.frame_base.offset.rot


class MobileEnv(DirectEnv):
    cfg: MobileEnvCfg

    def __init__(self, cfg: MobileEnvCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        ## Get scene assets
        self._imu_robot: Imu = self.scene["imu_robot"]



from srb.core.asset import AssetBaseCfg, Payload
from srb.core.sim import (
    CollisionPropertiesCfg,
    MassPropertiesCfg,
    MeshCollisionPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.path import SRB_ASSETS_DIR_SRB_OBJECT


class CargoBay(Payload):
    asset_cfg: AssetBaseCfg = AssetBaseCfg(
        prim_path="{ENV_REGEX_NS}/cargo_bay",
        spawn=UsdFileCfg(
            usd_path=(SRB_ASSETS_DIR_SRB_OBJECT.joinpath("cargo_bay.usdz").as_posix()),
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            mass_props=MassPropertiesCfg(density=1000.0),
        ),
    )



import subprocess

import pytest

from srb.utils import logging
from srb.utils.isaacsim import get_isaacsim_python


@pytest.mark.order(2)
def test_cli_ls():
    cmd = (
        get_isaacsim_python(),
        "-m",
        "srb",
        "ls",
        "--show_hidden",
    )

    try:
        process = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
        )

        if process.wait(timeout=600.0) != 0:
            logging.critical(f"Failed command: {' '.join(cmd)}")
            stdout, stderr = process.communicate()
            pytest.fail(f"Process failed\nSTDOUT:\n{stdout}\nSTDERR:\n{stderr}")
    except Exception as e:
        pytest.fail(f"Failed to start process\nException: {e}")



from __future__ import annotations

from functools import cached_property
from typing import ClassVar, Dict, Iterable, List, Sequence, Tuple, Type

from srb.core.asset.asset import Asset, AssetRegistry
from srb.core.asset.asset_type import AssetType
from srb.core.asset.object.object_type import ObjectType


class Object(Asset, asset_entrypoint=AssetType.OBJECT):
    def __init_subclass__(
        cls,
        object_entrypoint: ObjectType | None = None,
        object_metaclass: bool = False,
        asset_metaclass: bool = False,
        **kwargs,
    ):
        super().__init_subclass__(
            asset_metaclass=(
                asset_metaclass or object_entrypoint is not None or object_metaclass
            ),
            **kwargs,
        )
        if object_entrypoint is not None:
            assert isinstance(object_entrypoint, ObjectType), (
                f"Class '{cls.__name__}' is marked as an object entrypoint, but '{object_entrypoint}' is not a valid {ObjectType}"
            )
            assert object_entrypoint not in ObjectRegistry.base_types.keys(), (
                f"Class '{cls.__name__}' is marked as '{object_entrypoint}' object entrypoint, but it was already marked by '{ObjectRegistry.base_types[object_entrypoint].__name__}'"
            )
            ObjectRegistry.base_types[object_entrypoint] = cls
        elif object_metaclass:
            ObjectRegistry.meta_types.append(cls)
        elif not asset_metaclass:
            object_type = next(
                (
                    object_type
                    for object_type, base in ObjectRegistry.base_types.items()
                    if issubclass(cls, base)
                ),
                ObjectType.COMMON,
            )
            if object_type not in ObjectRegistry.registry.keys():
                ObjectRegistry.registry[object_type] = []
            else:
                assert cls.name() not in (
                    object.name() for object in ObjectRegistry.registry[object_type]
                ), (
                    f"Cannot register multiple sceneries with an identical name: '{cls.__module__}:{cls.__name__}' already exists as '{next(object for object in ObjectRegistry.registry[object_type] if cls.name() == object.name()).__module__}:{cls.__name__}'"
                )
            ObjectRegistry.registry[object_type].append(cls)

    @cached_property
    def object_type(self) -> ObjectType:
        for object_type, base in ObjectRegistry.base_types.items():
            if isinstance(self, base):
                return object_type
        return ObjectType.COMMON

    @classmethod
    def object_registry(cls) -> Sequence[Type[Object]]:
        return list(ObjectRegistry.values_inner())

    @classmethod
    def asset_registry(cls) -> Sequence[Type[Object]]:
        return AssetRegistry.registry.get(AssetType.OBJECT, [])  # type: ignore


class ObjectRegistry:
    registry: ClassVar[Dict[ObjectType, List[Type[Object]]]] = {}
    base_types: ClassVar[Dict[ObjectType, Type[Object]]] = {}
    meta_types: ClassVar[List[Type[Object]]] = []

    @classmethod
    def keys(cls) -> Iterable[ObjectType]:
        return cls.registry.keys()

    @classmethod
    def items(cls) -> Iterable[Tuple[ObjectType, Sequence[Type[Object]]]]:
        return cls.registry.items()

    @classmethod
    def values(cls) -> Iterable[Iterable[Type[Object]]]:
        return cls.registry.values()

    @classmethod
    def values_inner(cls) -> Iterable[Type[Object]]:
        return (object for sceneries in cls.registry.values() for object in sceneries)

    @classmethod
    def n_sceneries(cls) -> int:
        return sum(len(sceneries) for sceneries in cls.registry.values())

    @classmethod
    def registered_modules(cls) -> Iterable[str]:
        return {
            object.__module__
            for sceneries in cls.registry.values()
            for object in sceneries
        }

    @classmethod
    def registered_packages(cls) -> Iterable[str]:
        return {module.split(".", maxsplit=1)[0] for module in cls.registered_modules()}

    @classmethod
    def get_by_name(cls, name: str) -> Type[Object] | None:
        for object in cls.values_inner():
            if object.name() == name:
                return object
        return None



from .direct import DirectEnv, DirectEnvCfg  # noqa: F401
from .direct_marl import DirectMarlEnv, DirectMarlEnvCfg  # noqa: F401
from .env_cfg import BaseEnvCfg  # noqa: F401
from .event_cfg import BaseEventCfg  # noqa: F401
from .managed import ManagedEnv, ManagedEnvCfg  # noqa: F401
from .scene_cfg import BaseSceneCfg  # noqa: F401



from dataclasses import MISSING

import torch

from srb import assets
from srb.core.action import ThrustAction
from srb.core.asset import AssetVariant, OrbitalRobot, Scenery
from srb.core.domain import Domain
from srb.core.env import ViewerCfg
from srb.core.env.mobile.env import (
    MobileEnv,
    MobileEnvCfg,
    MobileEventCfg,
    MobileSceneCfg,
)
from srb.core.manager import EventTermCfg, SceneEntityCfg
from srb.core.mdp import reset_root_state_uniform
from srb.utils.cfg import configclass
from srb.utils.math import deg_to_rad


@configclass
class OrbitalSceneCfg(MobileSceneCfg):
    env_spacing: float = 4.0


@configclass
class OrbitalEventCfg(MobileEventCfg):
    randomize_robot_state: EventTermCfg = EventTermCfg(
        func=reset_root_state_uniform,
        mode="reset",
        params={
            "asset_cfg": SceneEntityCfg("robot"),
            "pose_range": {
                "x": (-0.2, 0.2),
                "y": (-0.2, 0.2),
                "z": (-0.2, 0.2),
                "roll": (-torch.pi, torch.pi),
                "pitch": (-torch.pi, torch.pi),
                "yaw": (-torch.pi, torch.pi),
            },
            "velocity_range": {
                "x": (-0.2, 0.2),
                "y": (-0.2, 0.2),
                "z": (-0.2, 0.2),
                "roll": (-deg_to_rad(10.0), deg_to_rad(10.0)),
                "pitch": (-deg_to_rad(10.0), deg_to_rad(10.0)),
                "yaw": (-deg_to_rad(10.0), deg_to_rad(10.0)),
            },
        },
    )


@configclass
class OrbitalEnvCfg(MobileEnvCfg):
    ## Scenario
    domain: Domain = Domain.ORBIT

    ## Assets
    robot: OrbitalRobot | AssetVariant = assets.Cubesat()
    _robot: OrbitalRobot = MISSING  # type: ignore
    scenery: Scenery | AssetVariant | None = None

    ## Scene
    scene: OrbitalSceneCfg = OrbitalSceneCfg()

    ## Events
    events: OrbitalEventCfg = OrbitalEventCfg()

    ## Time
    env_rate: float = 1.0 / 25.0
    agent_rate: float = 1.0 / 25.0

    ## Viewer
    viewer: ViewerCfg = ViewerCfg(
        eye=(10.0, -10.0, 10.0), lookat=(0.0, 0.0, 0.0), origin_type="env"
    )

    def __post_init__(self):
        super().__post_init__()


class OrbitalEnv(MobileEnv):
    cfg: OrbitalEnvCfg

    def __init__(self, cfg: OrbitalEnvCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        ## Check if thrust action is used (fuel tracking)
        self._thrust_action_term_key = next(
            filter(
                lambda term_key: isinstance(
                    self.action_manager._terms[term_key], ThrustAction
                ),
                self.action_manager._terms.keys(),
            ),
            None,
        )



import simforge_foundry

from srb.core.asset import Object, RigidObjectCfg
from srb.core.sim import (
    CollisionPropertiesCfg,
    MassPropertiesCfg,
    MeshCollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    SimforgeAssetCfg,
)


class LunarRock(Object):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/rock",
        spawn=SimforgeAssetCfg(
            assets=[simforge_foundry.MoonRock()],
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(),
            mass_props=MassPropertiesCfg(density=2000.0),
        ),
    )


class MarsRock(Object):
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/rock",
        spawn=SimforgeAssetCfg(
            assets=[simforge_foundry.MarsRock()],
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(),
            mass_props=MassPropertiesCfg(density=2000.0),
        ),
    )



try:
    import rclpy
except ImportError:
    from srb.utils.ros import enable_ros2_bridge

    enable_ros2_bridge()

import threading
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Dict, List, Sequence, Tuple, Type

import numpy
import rclpy
import torch
from builtin_interfaces.msg import Time
from geometry_msgs.msg import Quaternion, Transform, TransformStamped, Twist, Vector3
from rclpy.executors import MultiThreadedExecutor
from rclpy.node import Node
from rclpy.publisher import Publisher
from rclpy.qos import DurabilityPolicy, HistoryPolicy, QoSProfile, ReliabilityPolicy
from rclpy.service import Service
from rclpy.subscription import Subscription
from rosgraph_msgs.msg import Clock
from sensor_msgs.msg import CameraInfo, Image
from sensor_msgs.msg import Imu as ImuMsg
from sensor_msgs.msg import JointState, PointCloud2
from sensor_msgs.msg import PointField as ROSPointField
from std_msgs.msg import Bool as BoolMsg
from std_msgs.msg import (
    Float32,
    Float32MultiArray,
    Header,
    MultiArrayDimension,
    MultiArrayLayout,
)
from std_srvs.srv import Empty as EmptySrv
from tf2_ros import TransformBroadcaster
from tf2_ros.static_transform_broadcaster import StaticTransformBroadcaster

from srb.core.action import (
    ActionGroup,
    ActionTerm,
    ActionTermCfg,
    BinaryJointPositionAction,
    BinaryJointVelocityAction,
    BodyAccelerationAction,
    DifferentialInverseKinematicsAction,
    MulticopterBodyAccelerationAction,
    NonHolonomicAction,
    OperationalSpaceControllerAction,
    WheeledDriveAction,
)
from srb.core.asset import Articulation, RigidObject, RigidObjectCollection
from srb.core.manager import SimulationManager
from srb.core.sensor import Camera, Imu, RayCaster, RayCasterCamera
from srb.utils.camera import create_pointcloud_from_depth, create_pointcloud_from_rgbd
from srb.utils.math import subtract_frame_transforms

from .base import InterfaceBase

if TYPE_CHECKING:
    from srb._typing import AnyEnv


class RosInterface(InterfaceBase):
    __QOS_PROFILE = QoSProfile(
        reliability=ReliabilityPolicy.RELIABLE,
        durability=DurabilityPolicy.VOLATILE,
        history=HistoryPolicy.KEEP_LAST,
        depth=1,
    )
    __QOS_PROFILE_TF = QoSProfile(
        reliability=ReliabilityPolicy.RELIABLE,
        durability=DurabilityPolicy.TRANSIENT_LOCAL,
        history=HistoryPolicy.KEEP_LAST,
        depth=1,
    )

    def __init__(self, env: "AnyEnv", node: Node | None = None, *args, **kwargs):
        self._env: "AnyEnv" = env.unwrapped  # type: ignore
        self._num_envs = self._env.num_envs

        ## Initialize node
        if not node:
            rclpy.init()
            self._node = Node("sim", namespace="srb", start_parameter_services=False)  # type: ignore
        else:
            self._node = node

        ## Keep track of whether a first update has been called
        self._update_called = False

        ## Initialize queue for async executing of callbacks
        self._async_exec_queue: Dict[Callable, Dict[str, Any]] = {}

        ## Initialize action buffer
        assert self._env.action_space.shape is not None
        self._actions = torch.zeros(
            self._env.action_space.shape, dtype=torch.float32, device=self._env.device
        )

        ## Setup ROS interfaces
        # Pub: Clock
        self._clock_pub: Publisher = self._node.create_publisher(
            msg_type=Clock,
            topic="/clock",
            qos_profile=self.__QOS_PROFILE,
        )

        # Pub: Reward
        self._pub_reward: Sequence[Publisher] = tuple(
            self._node.create_publisher(
                Float32,
                f"env{i}/reward",
                qos_profile=self.__QOS_PROFILE,
            )
            for i in range(self._num_envs)
        )
        # Pub: Reward term
        self._pub_reward_term: Dict[str, Sequence[Publisher]] = {}

        # Pub: Terminated
        self._pub_terminated: Sequence[Publisher] = tuple(
            self._node.create_publisher(
                BoolMsg,
                f"env{i}/terminated",
                qos_profile=self.__QOS_PROFILE,
            )
            for i in range(self._num_envs)
        )
        # Pub: Truncated
        self._pub_truncated: Sequence[Publisher] = tuple(
            self._node.create_publisher(
                BoolMsg,
                f"env{i}/truncated",
                qos_profile=self.__QOS_PROFILE,
            )
            for i in range(self._num_envs)
        )

        # Srv: Reset all
        self._srv_reset_all: Service = self._node.create_service(
            EmptySrv, "envs/reset", self._create_cb_reset()
        )
        # Srv: Reset individual
        self._srv_reset: Sequence[Service] = tuple(
            self._node.create_service(
                EmptySrv, f"env{i}/reset", self._create_cb_reset(i)
            )
            for i in range(self._num_envs)
        )

        ## Pub: Transforms (all scene assets)
        self._tf_broadcaster = TransformBroadcaster(
            self._node, qos=self.__QOS_PROFILE_TF
        )
        self._tf_broadcaster_static = StaticTransformBroadcaster(
            self._node, qos=self.__QOS_PROFILE_TF
        )
        self._broadcast_transforms_static()

        ## Pub: Joint states (all scene articulations)
        self._pub_joint_states: Dict[str, Sequence[Publisher]] = {
            articulation_name: tuple(
                self._node.create_publisher(
                    JointState,
                    f"env{i}/{articulation_name}/joint_states",
                    qos_profile=self.__QOS_PROFILE,
                )
                for i in range(self._num_envs)
            )
            for articulation_name in self._env.scene._articulations.keys()
        }

        ## Pub: Sensors (all scene sensors)
        self._setup_sensor_interfaces()

        ## Spin up ROS executor
        if not node:
            self._executor = MultiThreadedExecutor(num_threads=2)
            self._executor.add_node(self._node)
            self._thread = threading.Thread(target=self._executor.spin)
            self._thread.daemon = True
            self._thread.start()

    def __del__(self):
        if hasattr(self, "_executor"):
            self._executor.shutdown()
            self._thread.join()
            self._node.destroy_node()

    @property
    def action(self) -> torch.Tensor:
        return self._actions

    def update(
        self,
        observation: Dict[str, torch.Tensor],
        reward: torch.Tensor,
        terminated: torch.Tensor,
        truncated: torch.Tensor,
        info: Dict[str, Any],
        action: torch.Tensor | None = None,
        *args,
        **kwargs,
    ):
        ## Perform delayed setup on first update
        if not self._update_called:
            self._update_called = True
            # Pub: Reward term
            if "reward_terms" in info:
                self._pub_reward_term = {
                    reward_term: tuple(
                        self._node.create_publisher(
                            Float32,
                            f"env{i}/reward/{reward_term}",
                            qos_profile=self.__QOS_PROFILE,
                        )
                        for i in range(self._num_envs)
                    )
                    for reward_term in info["reward_terms"]
                }

            # Pub: Actions (only if action subscription is not set up)
            if not hasattr(self, "_sub_action_term") and action is not None:
                self._pub_action = tuple(
                    self._node.create_publisher(
                        Float32MultiArray,
                        f"env{i}/action",
                        qos_profile=self.__QOS_PROFILE,
                    )
                    for i in range(self._num_envs)
                )

        ## Update clock
        time_s = self._env.common_step_counter * self._env.step_dt
        time_msg = Time(sec=int(time_s), nanosec=int((time_s % 1) * 1e9))
        self._clock_pub.publish(Clock(clock=time_msg))

        ## Publish per-env messages from the input
        for i in range(self._num_envs):
            self._pub_reward[i].publish(Float32(data=reward[i].item()))
            for reward_term, pubs in self._pub_reward_term.items():
                pubs[i].publish(
                    Float32(data=info["reward_terms"][reward_term][i].item())
                )
            self._pub_terminated[i].publish(BoolMsg(data=terminated[i].item()))
            self._pub_truncated[i].publish(BoolMsg(data=truncated[i].item()))

        ## Publish actions
        if action is not None and hasattr(self, "_pub_action"):
            layout = MultiArrayLayout(
                dim=[
                    MultiArrayDimension(
                        label="action",
                        size=action.shape[1],
                        stride=4 * action.shape[1],
                    )
                ],
                data_offset=0,
            )
            for i in range(self._num_envs):
                self._pub_action[i].publish(
                    Float32MultiArray(layout=layout, data=action[i].tolist())
                )

        ## Broadcast transforms (all scene assets)
        self._broadcast_transforms(time_msg)

        ## Publish joint states (all scene articulations)
        for articulation_name, pubs in self._pub_joint_states.items():
            articulation: Articulation = self._env.scene._articulations[
                articulation_name
            ]
            for i in range(self._num_envs):
                pubs[i].publish(
                    JointState(
                        header=Header(
                            stamp=time_msg,
                            frame_id=f"srb/env{i}/{articulation_name}",
                        ),
                        name=articulation.data.joint_names,
                        position=articulation.data.joint_pos[i],
                        velocity=articulation.data.joint_vel[i],
                        effort=articulation.data.applied_torque[i],
                    )
                )

        ## Publish sensor data (all scene sensors)
        self._publish_sensor_data(time_msg)

        ## Process async requests
        for request, kwargs in self._async_exec_queue.items():
            request(**kwargs)
        self._async_exec_queue.clear()

    ## Action ##

    def setup_action_sub(self):
        ## Action terms
        self._sub_action_term_all: Dict[str, Subscription] = {}
        self._sub_action_term: Dict[str, Sequence[Subscription]] = {}
        action_offset = 0
        for action_key, action_term in self._env.action_manager._terms.items():
            assert isinstance(action_term, ActionTerm)
            assert isinstance(action_term.cfg, ActionTermCfg)
            topic_base = (
                f"{action_term.cfg.asset_name}/{action_key.rsplit('__', 1)[-1]}"
            )
            msg_type, msg_extractor = self._map_action_term(action_term)
            action_dim = action_term.action_dim
            assert (
                msg_type is Float32MultiArray
                or len(msg_extractor(msg_type())) == action_dim
            )

            self._sub_action_term_all[action_key] = self._node.create_subscription(
                msg_type=msg_type,
                topic=f"envs/{topic_base}",
                callback=self._create_cb_action_any(
                    dim=action_dim,
                    key=action_key,
                    extractor=msg_extractor,
                    offset=action_offset,
                ),
                qos_profile=self.__QOS_PROFILE,
            )
            self._sub_action_term[action_key] = tuple(
                self._node.create_subscription(
                    msg_type=msg_type,
                    topic=f"env{i}/{topic_base}",
                    callback=self._create_cb_action_any(
                        key=action_key,
                        extractor=msg_extractor,
                        dim=action_dim,
                        offset=action_offset,
                        env_id=i,
                    ),
                    qos_profile=self.__QOS_PROFILE,
                )
                for i in range(self._num_envs)
            )

            action_offset += action_dim

        ## Standard actions
        if hasattr(self._env.cfg, "actions") and isinstance(
            self._env.cfg.actions, ActionGroup
        ):
            try:
                self._env.cfg.actions.map_cmd_to_action(torch.zeros(6), False)
                env_supports_direct_teleop = True
            except NotImplementedError:
                env_supports_direct_teleop = False
        else:
            env_supports_direct_teleop = False
        if env_supports_direct_teleop:
            self._actions_cmd_vel = torch.zeros(
                (self._num_envs, 6), dtype=torch.float32, device=self._env.device
            )
            self._actions_event = torch.zeros(
                (self._num_envs, 1), dtype=torch.float32, device=self._env.device
            )
            self._sub_action_cmd_vel_all: Subscription = self._node.create_subscription(
                Twist,
                "envs/action/cmd_vel",
                callback=self._create_cb_action_cmd_vel(),
                qos_profile=self.__QOS_PROFILE,
            )
            self._sub_action_cmd_vel: Sequence[Subscription] = tuple(
                self._node.create_subscription(
                    Twist,
                    f"env{i}/action/cmd_vel",
                    callback=self._create_cb_action_cmd_vel(env_id=i),
                    qos_profile=self.__QOS_PROFILE,
                )
                for i in range(self._num_envs)
            )
            self._sub_action_event_all: Subscription = self._node.create_subscription(
                BoolMsg,
                "envs/action/event",
                callback=self._create_cb_action_event(),
                qos_profile=self.__QOS_PROFILE,
            )
            self._sub_action_event: Sequence[Subscription] = tuple(
                self._node.create_subscription(
                    BoolMsg,
                    f"env{i}/action/event",
                    callback=self._create_cb_action_event(env_id=i),
                    qos_profile=self.__QOS_PROFILE,
                )
                for i in range(self._num_envs)
            )

    def _create_cb_action_any(
        self,
        key: str,
        extractor: Callable,
        dim: int,
        offset: int,
        env_id: int | None = None,
    ) -> Callable:
        start_idx = offset
        end_idx = offset + dim

        if env_id is not None:

            def _proto_cb(self, msg: Any):
                self._actions[env_id, start_idx:end_idx] = torch.tensor(
                    extractor(msg), dtype=torch.float32, device=self._env.device
                )

        else:

            def _proto_cb(self, msg: Any):
                self._actions[:, start_idx:end_idx] = torch.tensor(
                    extractor(msg), dtype=torch.float32, device=self._env.device
                )

        cb_name = f"__cb_action_{key}{env_id or ''}"
        setattr(self, cb_name, _proto_cb.__get__(self, self.__class__))
        return getattr(self, cb_name)

    @staticmethod
    def _map_action_term(action_term: ActionTerm) -> Tuple[Type, Callable]:
        if isinstance(
            action_term,
            (
                BodyAccelerationAction,
                DifferentialInverseKinematicsAction,
                OperationalSpaceControllerAction,
            ),
        ):
            return Twist, lambda msg: [
                msg.linear.x,
                msg.linear.y,
                msg.linear.z,
                msg.angular.x,
                msg.angular.y,
                msg.angular.z,
            ]

        if isinstance(action_term, (NonHolonomicAction, WheeledDriveAction)):
            return Twist, lambda msg: [msg.linear.x, msg.angular.z]

        if isinstance(action_term, MulticopterBodyAccelerationAction):
            return Twist, lambda msg: [
                msg.linear.x,
                msg.linear.y,
                msg.linear.z,
                msg.angular.z,
            ]

        if isinstance(
            action_term, (BinaryJointPositionAction, BinaryJointVelocityAction)
        ):
            return BoolMsg, lambda msg: [-1.0 if msg.data else 1.0]

        return Float32MultiArray, lambda msg: list(msg.data)

    def _create_cb_action_cmd_vel(self, env_id: int | None = None) -> Callable:
        if env_id is not None:

            def _proto_cb(self, msg: Twist):
                cmd_vel = torch.tensor(
                    [
                        msg.linear.x,
                        msg.linear.y,
                        msg.linear.z,
                        msg.angular.x,
                        msg.angular.y,
                        msg.angular.z,
                    ],
                    dtype=torch.float32,
                    device=self._env.device,
                )
                self._actions_cmd_vel[env_id] = cmd_vel
                self._actions[env_id] = self._env.cfg.actions.map_cmd_to_action(
                    cmd_vel, self._actions_event[env_id].bool()
                )

        else:

            def _proto_cb(self, msg: Twist):
                cmd_vel = torch.tensor(
                    [
                        msg.linear.x,
                        msg.linear.y,
                        msg.linear.z,
                        msg.angular.x,
                        msg.angular.y,
                        msg.angular.z,
                    ],
                    dtype=torch.float32,
                    device=self._env.device,
                )
                self._actions_cmd_vel = cmd_vel.unsqueeze(0).repeat(self._num_envs, 1)
                for i in range(self._num_envs):
                    self._actions[i] = self._env.cfg.actions.map_cmd_to_action(
                        cmd_vel, self._actions_event[i].bool()
                    )

        cb_name = f"__cb_action_cmd_vel{env_id or ''}"
        setattr(self, cb_name, _proto_cb.__get__(self, self.__class__))
        return getattr(self, cb_name)

    def _create_cb_action_event(self, env_id: int | None = None) -> Callable:
        if env_id is not None:

            def _proto_cb(self, msg: BoolMsg):
                event = msg.data
                self._actions_event[env_id] = event
                self._actions[env_id] = self._env.cfg.actions.map_cmd_to_action(
                    self._actions_cmd_vel[env_id], event
                )

        else:

            def _proto_cb(self, msg: BoolMsg):
                event = msg.data
                self._actions_event = (
                    torch.tensor([event], dtype=torch.float32, device=self._env.device)
                    .unsqueeze(0)
                    .repeat(self._num_envs, 1)
                )
                for i in range(self._num_envs):
                    self._actions[i] = self._env.cfg.actions.map_cmd_to_action(
                        self._actions_cmd_vel[i], event
                    )

        cb_name = f"__cb_action_event{env_id or ''}"
        setattr(self, cb_name, _proto_cb.__get__(self, self.__class__))
        return getattr(self, cb_name)

    ## Sensors ##

    def _setup_sensor_interfaces(self):
        self._pub_sensors: Dict[str, Dict[str, Sequence[Publisher]]] = {}

        for sensor_name, sensor in self._env.scene._sensors.items():
            self._pub_sensors[sensor_name] = {}

            if isinstance(sensor, (Camera, RayCasterCamera)):
                self._pub_sensors[sensor_name]["camera_info"] = tuple(
                    self._node.create_publisher(
                        CameraInfo,
                        f"env{i}/{sensor_name}/camera_info",
                        qos_profile=self.__QOS_PROFILE,
                    )
                    for i in range(self._num_envs)
                )

                has_depth_streams = False
                for data_type in sensor.cfg.data_types:
                    if data_type in (
                        "depth",
                        "distance_to_camera",
                        "distance_to_image_plane",
                    ):
                        has_depth_streams = True
                    image_name = (
                        "image_depth"
                        if data_type == "distance_to_image_plane"
                        else f"image_{data_type}"
                    )
                    self._pub_sensors[sensor_name][image_name] = tuple(
                        self._node.create_publisher(
                            Image,
                            f"env{i}/{sensor_name}/{image_name}",
                            qos_profile=self.__QOS_PROFILE,
                        )
                        for i in range(self._num_envs)
                    )

                if has_depth_streams:
                    self._pub_sensors[sensor_name]["pointcloud"] = tuple(
                        self._node.create_publisher(
                            PointCloud2,
                            f"env{i}/{sensor_name}/pointcloud",
                            qos_profile=self.__QOS_PROFILE,
                        )
                        for i in range(self._num_envs)
                    )

            elif isinstance(sensor, RayCaster):
                self._pub_sensors[sensor_name]["pointcloud"] = tuple(
                    self._node.create_publisher(
                        PointCloud2,
                        f"env{i}/{sensor_name}/pointcloud",
                        qos_profile=self.__QOS_PROFILE,
                    )
                    for i in range(self._num_envs)
                )

            elif isinstance(sensor, Imu):
                self._pub_sensors[sensor_name]["imu"] = tuple(
                    self._node.create_publisher(
                        ImuMsg,
                        f"env{i}/{sensor_name}",
                        qos_profile=self.__QOS_PROFILE,
                    )
                    for i in range(self._num_envs)
                )

    def _publish_sensor_data(self, time_msg: Time):
        for sensor_name, publishers in self._pub_sensors.items():
            sensor = self._env.scene._sensors[sensor_name]

            ## Cameras
            if isinstance(sensor, (Camera, RayCasterCamera)):
                # Camera info
                for i in range(self._num_envs):
                    camera_info_msg = CameraInfo()
                    camera_info_msg.header = Header(
                        stamp=time_msg, frame_id=f"srb/env{i}/{sensor_name}"
                    )
                    camera_info_msg.height = sensor.image_shape[0]
                    camera_info_msg.width = sensor.image_shape[1]
                    camera_info_msg.distortion_model = "plumb_bob"
                    camera_info_msg.d = [0.0, 0.0, 0.0, 0.0, 0.0]  # No distortion
                    intrinsic_matrix = sensor.data.intrinsic_matrices[i]
                    camera_info_msg.k = [
                        intrinsic_matrix[0, 0].item(),
                        intrinsic_matrix[0, 1].item(),
                        intrinsic_matrix[0, 2].item(),
                        intrinsic_matrix[1, 0].item(),
                        intrinsic_matrix[1, 1].item(),
                        intrinsic_matrix[1, 2].item(),
                        intrinsic_matrix[2, 0].item(),
                        intrinsic_matrix[2, 1].item(),
                        intrinsic_matrix[2, 2].item(),
                    ]
                    publishers["camera_info"][i].publish(camera_info_msg)

                # Images
                for data_type in sensor.data.output.keys():
                    image_name = (
                        "image_depth"
                        if data_type == "distance_to_image_plane"
                        else f"image_{data_type}"
                    )
                    img_data_all = sensor.data.output[data_type].cpu().numpy()
                    for i in range(self._num_envs):
                        img_data = img_data_all[i]

                        # Create image message
                        image_msg = Image()
                        image_msg.header = Header(
                            stamp=time_msg, frame_id=f"srb/env{i}/{sensor_name}"
                        )
                        image_msg.height = img_data.shape[0]
                        image_msg.width = img_data.shape[1]

                        # Handle different image types
                        if data_type == "rgb" or img_data.shape[2] == 3:
                            image_msg.encoding = "rgb8"
                            image_msg.is_bigendian = False
                            image_msg.step = 3 * img_data.shape[1]
                            if img_data.dtype != numpy.uint8:
                                img_data = (255.0 * img_data).astype(numpy.uint8)
                            image_msg.data = img_data.tobytes()

                        elif data_type == "rgba" or img_data.shape[2] == 4:
                            image_msg.encoding = "rgba8"
                            image_msg.is_bigendian = False
                            image_msg.step = 4 * img_data.shape[1]
                            if img_data.dtype != numpy.uint8:
                                img_data = (255.0 * img_data).astype(numpy.uint8)
                            image_msg.data = img_data.tobytes()

                        elif data_type in (
                            "depth",
                            "distance_to_camera",
                            "distance_to_image_plane",
                        ):
                            image_msg.encoding = "32FC1"
                            image_msg.is_bigendian = False
                            image_msg.step = 4 * img_data.shape[1]
                            image_msg.data = img_data.astype(numpy.float32).tobytes()

                        elif img_data.shape[2] == 1:
                            image_msg.encoding = "mono8"
                            image_msg.is_bigendian = False
                            image_msg.step = img_data.shape[1]
                            if img_data.dtype != numpy.uint8:
                                img_data = (255.0 * img_data).astype(numpy.uint8)
                            image_msg.data = img_data.tobytes()

                        else:
                            # Unsupported format
                            break

                        publishers[image_name][i].publish(image_msg)

                # Pointcloud
                depth_type = None
                for depth_option in (
                    "depth",
                    "distance_to_image_plane",
                    "distance_to_camera",
                ):
                    if depth_option in sensor.data.output:
                        depth_type = depth_option
                        break
                if depth_type:
                    depth_data = sensor.data.output[depth_type]

                    if "rgb" in sensor.data.output.keys():
                        rgb_data = sensor.data.output["rgb"].cpu().numpy()
                    elif "rgba" in sensor.data.output.keys():
                        rgb_data = sensor.data.output["rgba"][..., :3].cpu().numpy()
                    else:
                        rgb_data = None

                    for i in range(self._num_envs):
                        if rgb_data is not None:
                            points, colors = create_pointcloud_from_rgbd(
                                intrinsic_matrix=sensor.data.intrinsic_matrices[i],
                                depth=depth_data[i],
                                rgb=rgb_data[i],
                                normalize_rgb=True,
                                device=self._env.device,
                            )
                            colors = colors.cpu().numpy().astype(numpy.float32)  # type: ignore
                        else:
                            points = create_pointcloud_from_depth(
                                intrinsic_matrix=sensor.data.intrinsic_matrices[i],
                                depth=depth_data[i],
                                device=self._env.device,
                            )
                        points = points.cpu().numpy().astype(numpy.float32)  # type: ignore

                        # Create PointCloud2 message
                        pointcloud_msg = PointCloud2()
                        pointcloud_msg.header = Header(
                            stamp=time_msg, frame_id=f"srb/env{i}/{sensor_name}"
                        )
                        pointcloud_msg.height = 1
                        pointcloud_msg.width = points.shape[0]  # type: ignore

                        # Define fields
                        fields = [
                            ROSPointField(
                                name="x",
                                offset=0,
                                datatype=ROSPointField.FLOAT32,
                                count=1,
                            ),
                            ROSPointField(
                                name="y",
                                offset=4,
                                datatype=ROSPointField.FLOAT32,
                                count=1,
                            ),
                            ROSPointField(
                                name="z",
                                offset=8,
                                datatype=ROSPointField.FLOAT32,
                                count=1,
                            ),
                        ]
                        if rgb_data is not None:
                            fields.extend(
                                [
                                    ROSPointField(
                                        name="r",
                                        offset=12,
                                        datatype=ROSPointField.FLOAT32,
                                        count=1,
                                    ),
                                    ROSPointField(
                                        name="g",
                                        offset=16,
                                        datatype=ROSPointField.FLOAT32,
                                        count=1,
                                    ),
                                    ROSPointField(
                                        name="b",
                                        offset=20,
                                        datatype=ROSPointField.FLOAT32,
                                        count=1,
                                    ),
                                ]
                            )
                            pointcloud_data = numpy.hstack([points, colors])
                            point_step = 24
                        else:
                            pointcloud_data = points
                            point_step = 12

                        pointcloud_msg.fields = fields
                        pointcloud_msg.is_bigendian = False
                        pointcloud_msg.point_step = point_step
                        pointcloud_msg.row_step = (
                            pointcloud_msg.point_step * pointcloud_msg.width
                        )
                        pointcloud_msg.is_dense = True
                        pointcloud_msg.data = pointcloud_data.tobytes()

                        publishers["pointcloud"][i].publish(pointcloud_msg)

            ## RayCaster (non-camera)
            elif isinstance(sensor, RayCaster):
                ray_hits_all = (
                    sensor.data.ray_hits_w.cpu().numpy().astype(numpy.float32)
                )
                for i in range(self._num_envs):
                    ray_hits = ray_hits_all[i]

                    pointcloud_msg = PointCloud2()
                    pointcloud_msg.header = Header(
                        stamp=time_msg, frame_id=f"srb/env{i}/{sensor_name}"
                    )
                    pointcloud_msg.height = 1
                    pointcloud_msg.width = ray_hits.shape[0]

                    # Define fields
                    fields = [
                        ROSPointField(
                            name="x", offset=0, datatype=ROSPointField.FLOAT32, count=1
                        ),
                        ROSPointField(
                            name="y", offset=4, datatype=ROSPointField.FLOAT32, count=1
                        ),
                        ROSPointField(
                            name="z", offset=8, datatype=ROSPointField.FLOAT32, count=1
                        ),
                    ]

                    pointcloud_msg.fields = fields
                    pointcloud_msg.is_bigendian = False
                    pointcloud_msg.point_step = 12
                    pointcloud_msg.row_step = (
                        pointcloud_msg.point_step * pointcloud_msg.width
                    )
                    pointcloud_msg.is_dense = True
                    pointcloud_msg.data = ray_hits.tobytes()

                    publishers["pointcloud"][i].publish(pointcloud_msg)

            # IMU
            elif isinstance(sensor, Imu):
                lin_acc = sensor.data.lin_acc_b
                ang_vel = sensor.data.ang_vel_b
                for i in range(self._num_envs):
                    imu_msg = ImuMsg()
                    imu_msg.header = Header(
                        stamp=time_msg, frame_id=f"srb/env{i}/{sensor_name}"
                    )
                    imu_msg.linear_acceleration.x = lin_acc[i, 0].item()
                    imu_msg.linear_acceleration.y = lin_acc[i, 1].item()
                    imu_msg.linear_acceleration.z = lin_acc[i, 2].item()
                    imu_msg.angular_velocity.x = ang_vel[i, 0].item()
                    imu_msg.angular_velocity.y = ang_vel[i, 1].item()
                    imu_msg.angular_velocity.z = ang_vel[i, 2].item()

                    publishers["imu"][i].publish(imu_msg)

    ## Transforms ##

    def _broadcast_transforms(self, time_msg: Time):
        transforms: List[TransformStamped] = []
        for asset_name, asset in (
            self._env.scene._rigid_objects.items()
            | self._env.scene._articulations.items()
        ):
            assert isinstance(asset, (RigidObject, Articulation))
            root_pos = asset.data.root_pos_w - self._env.scene.env_origins
            transforms.extend(
                [
                    TransformStamped(
                        header=Header(
                            stamp=time_msg,
                            frame_id=f"srb/env{i}",
                        ),
                        child_frame_id=f"srb/env{i}/{asset_name}",
                        transform=Transform(
                            translation=Vector3(
                                x=root_pos[i, 0].item(),
                                y=root_pos[i, 1].item(),
                                z=root_pos[i, 2].item(),
                            ),
                            rotation=Quaternion(
                                w=asset.data.root_quat_w[i, 0].item(),
                                x=asset.data.root_quat_w[i, 1].item(),
                                y=asset.data.root_quat_w[i, 2].item(),
                                z=asset.data.root_quat_w[i, 3].item(),
                            ),
                        ),
                    )
                    for i in range(self._num_envs)
                ]
            )
        for asset_name, asset in self._env.scene._sensors.items():
            if isinstance(asset, (Camera, RayCasterCamera)):
                root_pos = asset.data.pos_w - self._env.scene.env_origins
                root_quat = asset.data.quat_w_ros
            elif isinstance(asset, (RayCaster, Imu)):
                root_pos = asset.data.pos_w - self._env.scene.env_origins
                root_quat = asset.data.quat_w
            else:
                continue

            transforms.extend(
                [
                    TransformStamped(
                        header=Header(
                            stamp=time_msg,
                            frame_id=f"srb/env{i}",
                        ),
                        child_frame_id=f"srb/env{i}/{asset_name}",
                        transform=Transform(
                            translation=Vector3(
                                x=root_pos[i, 0].item(),
                                y=root_pos[i, 1].item(),
                                z=root_pos[i, 2].item(),
                            ),
                            rotation=Quaternion(
                                w=root_quat[i, 0].item(),  # type: ignore
                                x=root_quat[i, 1].item(),  # type: ignore
                                y=root_quat[i, 2].item(),  # type: ignore
                                z=root_quat[i, 3].item(),  # type: ignore
                            ),
                        ),
                    )
                    for i in range(self._num_envs)
                ]
            )

        for asset_name, asset in self._env.scene._rigid_object_collections.items():
            assert isinstance(asset, RigidObjectCollection)
            object_pos = (
                asset.data.object_pos_w
                - self._env.scene.env_origins.unsqueeze(1).repeat(
                    1, asset.num_objects, 1
                )
            )
            transforms.extend(
                [
                    TransformStamped(
                        header=Header(
                            stamp=time_msg,
                            frame_id=f"srb/env{i}",
                        ),
                        child_frame_id=f"srb/env{i}/{object_name}",
                        transform=Transform(
                            translation=Vector3(
                                x=object_pos[i, object_id, 0].item(),
                                y=object_pos[i, object_id, 1].item(),
                                z=object_pos[i, object_id, 2].item(),
                            ),
                            rotation=Quaternion(
                                w=asset.data.object_quat_w[i, object_id, 0].item(),
                                x=asset.data.object_quat_w[i, object_id, 1].item(),
                                y=asset.data.object_quat_w[i, object_id, 2].item(),
                                z=asset.data.object_quat_w[i, object_id, 3].item(),
                            ),
                        ),
                    )
                    for i in range(self._num_envs)
                    for object_id, object_name in enumerate(asset.object_names)
                ]
            )
        for asset_name, asset in self._env.scene._articulations.items():
            assert isinstance(asset, Articulation)
            body_pos, body_quat = subtract_frame_transforms(
                asset.data.root_pos_w.unsqueeze(1).repeat(1, asset.num_bodies, 1),
                asset.data.root_quat_w.unsqueeze(1).repeat(1, asset.num_bodies, 1),
                asset.data.body_pos_w,
                asset.data.body_quat_w,
            )
            transforms.extend(
                [
                    TransformStamped(
                        header=Header(
                            stamp=time_msg,
                            frame_id=f"srb/env{i}/{asset_name}",
                        ),
                        child_frame_id=f"srb/env{i}/{asset_name}/{body_name}",
                        transform=Transform(
                            translation=Vector3(
                                x=body_pos[i, body_id, 0].item(),
                                y=body_pos[i, body_id, 1].item(),
                                z=body_pos[i, body_id, 2].item(),
                            ),
                            rotation=Quaternion(
                                w=body_quat[i, body_id, 0].item(),
                                x=body_quat[i, body_id, 1].item(),
                                y=body_quat[i, body_id, 2].item(),
                                z=body_quat[i, body_id, 3].item(),
                            ),
                        ),
                    )
                    for i in range(self._num_envs)
                    for (
                        body_name,
                        body_id,
                    ) in asset.root_physx_view.shared_metatype.link_indices.items()
                ]
            )
        self._tf_broadcaster.sendTransform(transforms)

    def _broadcast_transforms_static(self):
        static_transforms = [
            # World -> Map (alias for simplicity)
            TransformStamped(
                header=Header(
                    stamp=Time(sec=0, nanosec=0),
                    frame_id="world",
                ),
                child_frame_id="map",
                transform=Transform(
                    translation=Vector3(x=0.0, y=0.0, z=0.0),
                    rotation=Quaternion(w=1.0, x=0.0, y=0.0, z=0.0),
                ),
            ),
            # World -> SRB World
            TransformStamped(
                header=Header(
                    stamp=Time(sec=0, nanosec=0),
                    frame_id="world",
                ),
                child_frame_id="srb/world",
                transform=Transform(
                    translation=Vector3(x=0.0, y=0.0, z=0.0),
                    rotation=Quaternion(w=1.0, x=0.0, y=0.0, z=0.0),
                ),
            ),
            # SRB World -> Envs
            *(
                TransformStamped(
                    header=Header(
                        stamp=Time(sec=0, nanosec=0),
                        frame_id="srb/world",
                    ),
                    child_frame_id=f"srb/env{i}",
                    transform=Transform(
                        translation=Vector3(
                            x=self._env.scene.env_origins[i, 0].item(),
                            y=self._env.scene.env_origins[i, 1].item(),
                            z=self._env.scene.env_origins[i, 2].item(),
                        ),
                        rotation=Quaternion(
                            w=1.0,
                            x=0.0,
                            y=0.0,
                            z=0.0,
                        ),
                    ),
                )
                for i in range(self._num_envs)
            ),
        ]
        self._tf_broadcaster_static.sendTransform(static_transforms)

    ## Reset ##

    def _create_cb_reset(self, env_id: int | None = None) -> Callable:
        if env_id is not None:

            def _proto_cb(self, request: EmptySrv.Request, response: EmptySrv.Response):
                if self._async_reset in self._async_exec_queue:
                    if "env_ids" in self._async_exec_queue[self._async_reset]:
                        self._async_exec_queue[self._async_reset]["env_ids"].add(env_id)
                else:
                    self._async_exec_queue[self._async_reset] = {"env_ids": {env_id}}

                return response

        else:

            def _proto_cb(self, request: EmptySrv.Request, response: EmptySrv.Response):
                self._async_exec_queue[self._async_reset] = {}
                return response

        cb_name = f"__cb_reset{env_id or ''}"
        setattr(self, cb_name, _proto_cb.__get__(self, self.__class__))
        return getattr(self, cb_name)

    def _async_reset(self, env_ids: Sequence[int] | None = None):
        if env_ids:
            ids = torch.tensor(list(env_ids), dtype=torch.int, device=self._env.device)
            self._actions[ids].zero_()
            self._env._reset_idx(ids)  # type: ignore

            self._env.scene.write_data_to_sim()
            self._env.sim.forward()

            if self._env.sim.has_rtx_sensors():
                if self._env.cfg.rerender_on_reset:
                    self._env.sim.render()
                if self._env.cfg.wait_for_textures:
                    while SimulationManager.assets_loading():
                        self._env.sim.render()
        else:
            self._actions.zero_()
            self._env.reset()



from .cfg import VisualsCfg  # noqa: F401
from .rtx_post import (  # noqa: F401
    auto_exposure,
    chromatic_aberration,
    depth_of_field,
    fog,
    lens_flare,
    motion_blur,
    reshade,
    tv_noise,
)
